/*
(popen "command" "type")
  Works like man 3 popen.
  Returns a stream.

(pclose stream)
  Closes streams opened by popen.
  Returns exit status of command.
*/

#define CSTREAM_POPEN 0xb0b34

static int popen_reader(void **data, const int type)
{
	FILE *fp = (FILE *) *data;

	int c = fgetc(fp);
	if(c == EOF) return -1;
	return c;
}

static int popen_writer(int ch, void **data, const int type)
{
	FILE *fp = (FILE *) *data;

	if(fputc(ch, fp) == EOF) return -1;
	return 0;
}

static size_t popen_readmore(void **data, const int type, void *buf, const size_t len)
{
	FILE *fp = (FILE *) *data;

	return fread(buf, 1, len, fp);
}

static size_t popen_writemore(void **data, const int type, void *buf, const size_t len)
{
	FILE *fp = (FILE *) *data;

	return fwrite(buf, 1, len, fp);
}

static size_t popen_vprintf(void **data, const int type, const char *fmt, va_list ap)
{
	FILE *fp = (FILE *) *data;

	return vfprintf(fp, fmt, ap);
}

static void popen_destroy(stt_obj_t *o)
{
	FILE *fp = (FILE *) o->d.stream.s.c.data;
	if(fp) pclose(fp);
}

/* builtin: (popen cmd type) */
static STT_BUILTIN(sb_popen)
{
	char *cmd, *type;
	stt_obj_t *str;
	FILE *fp;

	STT_L_STRING(ctx->stt, cmd);
	STT_L_STRING(ctx->stt, type);

	fp = popen(cmd, type);
	if(!fp) return stt_create_error_with_obj(ctx->stt, STT_SE_IO, "popen failed", stt_create_string(ctx->stt, strerror(errno)));

	str = stt_create_stream(ctx->stt, STT_STREAM_C);
	str->d.stream.s.c.type = CSTREAM_POPEN;
	str->d.stream.s.c.data = (void *) fp;
	str->d.stream.s.c.reader = popen_reader;
	str->d.stream.s.c.writer = popen_writer;
	str->d.stream.s.c.readmore = popen_readmore;
	str->d.stream.s.c.writemore = popen_writemore;
	str->d.stream.s.c.vprintf = popen_vprintf;
	str->d.stream.s.c.destroy = popen_destroy;

	return str;
}

/* builtin: (pclose stream) */
static STT_BUILTIN(sb_pclose)
{
	stt_obj_t *str;
	int r;

	STT_ARG_STREAM(ctx->stt, str);

	if(str->d.stream.type != STT_STREAM_C || str->d.stream.s.c.type != CSTREAM_POPEN)
		return stt_create_error(ctx->stt, STT_SE_TYPE, "pclose called on non-popen stream");
	if(!str->d.stream.s.c.data)
		return stt_create_error(ctx->stt, STT_SE_IO, "popen stream already closed");

	r = pclose((FILE *) str->d.stream.s.c.data);
	str->d.stream.s.c.data = NULL;

	return STT_INT(ctx->stt, WEXITSTATUS(r));
}
