#ifndef _ST_MEMORY_H
#define _ST_MEMORY_H

#include <estutter/types.h>
#include <estutter/stt_config.h>

typedef int stt_s_mark;

/* wrapper for calloc with error handling */
STT_INLINE void *stt_calloc(size_t nmemb, size_t size);
/* wrapper for malloc with error handling */
STT_INLINE void *stt_malloc(size_t size);
/* wrapper for realloc with error handling */
STT_INLINE void *stt_realloc(void *ptr, size_t size);

/* try to take an element of the recycle list */
stt_obj_t *stt_obj_respawn(stt_ctx_t *stt);

/* variable */
/* typedef struct stt_var_t s_var; // forward-declared in types.h */

struct stt_var_s {
	char *name; /* name */
	unsigned long sdbm; /* name sdbm */
	unsigned long id; /* symbol id */
	stt_obj_t *value; /* value */
	stt_var_t *next; /* next variable in this chain */
};

/* variable context */

/* forward declaration in types.h: */
/* typedef structst t_varctx_s varctx; */

struct stt_varctx_s {
	stt_ctx_t *stt;
	stt_varctx_t *parent; /* parent context */
	long ns; /* number of slots in the variable hash (power of two) */
	long nsm; /* ns - 1 (mask) */
	stt_var_t **hash; /* variable hash */
	unsigned long id; /* stt_varctx_t ID */
	int weak; /* 1 if weak stt_varctx_t (local set doesn't create a new bind) */

	stt_obj_t *ref; /* stt_varctx_t object that references this ctx */
	unsigned long int mark; /* used by GC */
	int defer; /* use a deferred hash (for closures) */

	void *user_data; /* reserved for caller's use */
};

/* used list of protected stt_varctx_t roots */

typedef struct stt_protroot_s stt_protroot_t;
struct stt_protroot_s {
	stt_varctx_t *root;
	stt_protroot_t *next;
};

/* protect a stt_varctx_t root (scan for variables in it) */
void stt_gc_protect_root(stt_varctx_t *ctx);
/* unprotect a stt_varctx_t root (no longer scan for variables in it) */
void stt_gc_unprotect_root(stt_varctx_t *ctx);

/* add an object to the root set */
void stt_gc_protect_obj(stt_obj_t *obj);

/* remove an object from the root set */
void stt_gc_unprotect_obj(stt_obj_t *obj);

/* create variable context */
stt_varctx_t *stt_varctx_create(stt_ctx_t *stt, stt_varctx_t *parent, const long size);

/* compute hash value for a string */
unsigned long stt_sdbm(const char *str);

/* compute hash value for a string, ignoring case */
unsigned long stt_sdbmc(const char *str);

/* get hash stt_var_t of a symbol */
stt_var_t *stt_get_symbolsv(stt_ctx_t *stt, const char *name, unsigned long sdbmv);

/* get symbol name by symbol id */
char *stt_get_symbolname(stt_ctx_t *stt, unsigned long sdbmv, unsigned long id);

/* get unique symbol id of a symbol */
unsigned long stt_get_symbolid(stt_ctx_t *stt, const char *name, unsigned long sdbmv);

/* utility function: get symbol id by string */
unsigned long stt_symbolid(stt_ctx_t *stt, char *name);

/* store a variable inside a ctx */
void stt_varctx_store(stt_varctx_t *ctx, char *name, stt_obj_t *val);
void stt_varctx_store_sdbm(stt_varctx_t *ctx, char *name, stt_obj_t *val, unsigned long sdbmv);
void stt_varctx_store_s(stt_varctx_t *ctx, unsigned long sdbmv, unsigned long id, stt_obj_t *val);

/* get variable from a ctx (local only)
 * returns NULL if variable doesn't exist */
stt_var_t *stt_varctx_retr(stt_varctx_t *ctx, char *name);
stt_var_t *stt_varctx_retr_sdbm(stt_varctx_t *ctx, char *name, unsigned long sdbmv);
stt_var_t *stt_varctx_retr_s(stt_varctx_t *ctx, unsigned long sdbmv, unsigned long id);

/* get variable from a ctx (with parent fallback)
 * returns NULL if variable doesn't exist */
stt_var_t *stt_varctx_get(stt_varctx_t *ctx, char *name);
stt_var_t *stt_varctx_get_sdbm(stt_varctx_t *ctx, char *name, unsigned long sdbmv);
stt_var_t *stt_varctx_get_s(stt_varctx_t *ctx, unsigned long sdbmv, unsigned long id);

/* set/store a variable inside a ctx
 * (updates if the var already exists) */
void stt_varctx_set(stt_varctx_t *ctx, const char *name, stt_obj_t *val);
void stt_varctx_set_sdbm(stt_varctx_t *ctx, char *name, stt_obj_t *val, unsigned long sdbmv);
void stt_varctx_set_s(stt_varctx_t *ctx, unsigned long sdbmv, unsigned long id, stt_obj_t *val);

/* dump contents of stt_varctx_t to stderr */
stt_obj_t *stt_varctx_dump(stt_varctx_t *ctx);

/* free stt_varctx_t */
void stt_varctx_free(stt_varctx_t *ctx);

/* add object to protection list */
STT_INLINE void stt_gc_protlist_add(stt_ctx_t *stt, stt_obj_t *obj);

/* record allocated object for GC */
void stt_gc_record(stt_ctx_t *stt, stt_obj_t *obj);

/* remove last allocated object from GC records (in case you want to free it yourself) */
void stt_gc_unrecord(stt_ctx_t *stt);

/* checks whether the link should be followed
 * (not if the target object has the link in its stt_weakptr_t list 
 * invokes gc_mark_obj as necessary */
STT_INLINE void stt_gc_follow(stt_ctx_t *stt, stt_obj_t **obj);

/* free all data associated with object */
STT_INLINE void stt_gc_harvest(stt_ctx_t *stt, stt_obj_t *obj);

/* mark variables reachable from object */
STT_INLINE void stt_gc_mark_obj(stt_ctx_t *stt, stt_obj_t *obj);

/* mark variables in context */
void stt_gc_mark(stt_varctx_t *ctx);

/* do garbage collection. returns number of objects freed */
int stt_gc_sweep(stt_ctx_t *stt, stt_varctx_t *ctx);

/* get current mark in protection stack */
stt_s_mark stt_gc_prot_mark(stt_ctx_t *stt);

/* free up to mark in protection stack, except for object ret */
void stt_gc_prot_free(stt_ctx_t *stt, stt_s_mark mark, stt_obj_t *ret);

/* mark misc. pointer for free() */
void stt_mark_miscfree(stt_ctx_t *stt, void *ptr);

/* free pointers in miscfree list */
void stt_free_miscfree(stt_ctx_t *stt);

/* initialize Stutter (called by resetparser if necessary) */
void stt_init(stt_ctx_t *stt);

/* shut down Stutter */
void stt_uninit(stt_ctx_t *stt);

/* newhash */
unsigned long stt_hash(unsigned char *k, unsigned long int length, unsigned long int initval);

#endif
