#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include <math.h>
#include <stdarg.h>
#include <errno.h>
#include "types.h"
#include "memory.h"
#include "eval.h"
#include "builtins.h"
#include "compat.h"

static stt_obj_t *cached_nil = NULL, *cached_t = NULL, *cached_eof = NULL;

static int cookie_type_cnt = 0; /* first free cookie type id */

/******************************************** OBJECT CREATION ***/

/* creates and returns a new stt_obj_t, zeroed */
stt_obj_t *stt_alloc_obj(stt_ctx_t *stt)
{
	stt_obj_t *o;

	o = stt_obj_respawn(stt);
	if(!o) o = (stt_obj_t *) stt_calloc(1, sizeof(stt_obj_t));

	o->file = stt->lastcall_file;
	o->line = stt->lastcall_line;
	o->serial = stt->next_serial ++;

	stt_gc_record(stt, o);

	return o;
}

/* creates a new stt_obj_t without updating the records of the GC */
stt_obj_t *stt_alloc_ungc_obj(stt_ctx_t *stt)
{
	stt_obj_t *o;
	o = (stt_obj_t *) stt_calloc(1, sizeof(stt_obj_t));
	o->file = stt->file;
	o->line = stt->parsed_line;

	return o;
}

/* create a numeric atom */
stt_obj_t *stt_create_num(stt_ctx_t *stt, const double val)
{
	stt_obj_t *o = stt_alloc_obj(stt);
	o->type = STT_ST_NUM;
	o->d.num.value = val;
	return o;
}

/* create an integer atom */
stt_obj_t *stt_create_int(stt_ctx_t *stt, const long val)
{
#ifdef DISABLE_INTCACHE
	stt_obj_t *o = stt_alloc_obj(stt);
	o->type = STT_ST_INT;
	o->d.integer.value = val;
	return o;
#else
	stt_obj_t *o;
	if(val >= STT_REUSABLE_INT_LIMIT || val < 0) {
		o = stt_alloc_obj(stt);
		o->type = STT_ST_INT;
		o->d.integer.value = val;
		return o;
	} else {
		return stt->reusable_ints[val];
	}
#endif
}

/* create a character atom */
stt_obj_t *stt_create_char(stt_ctx_t *stt, const unsigned char ch)
{
#ifdef DISABLE_CHARCACHE
	stt_obj_t *o = stt_alloc_obj(stt);
	o->type = STT_ST_CHAR;
	o->d.ch.ch = val;
	return o;
#else
	return stt->reusable_chars[ch];
#endif
}

/* create a symbolic atom */
stt_obj_t *stt_create_symbol(stt_ctx_t *stt, char *val)
{
	unsigned long sdbmv = stt_sdbmc(val);
	stt_var_t *sv = stt_get_symbolsv(stt, val, sdbmv);
	stt_obj_t *o;

	if(sv->value) return sv->value;

	o = stt_alloc_obj(stt);
	o->type = STT_ST_SYMBOL;
	o->d.symbol.name = stt_strdup(sv->name);
	o->d.symbol.sdbm = sdbmv;
	o->d.symbol.id = sv->id;
	sv->value = o;

	return o;
}

/* create a guaranteed new symbolic atom */
stt_obj_t *stt_create_new_symbol(stt_ctx_t *stt, char *val)
{
	stt_obj_t *o = stt_alloc_obj(stt);
	char *c;
	o->type = STT_ST_SYMBOL;
	o->d.symbol.name = stt_strdup(val);
	c = o->d.symbol.name;
	for(; *c; c++) *c = toupper(*c);
	o->d.symbol.sdbm = stt_sdbm(o->d.symbol.name);
	o->d.symbol.id = stt_get_symbolid(stt, o->d.symbol.name, o->d.symbol.sdbm);
	return o;
}

/* create a keyword atom */
stt_obj_t *stt_create_keyword(stt_ctx_t *stt, char *val)
{
	stt_obj_t *o = stt_create_symbol(stt, val);
	o->d.symbol.colon = 1;
	return o;
}

/* create a string */
stt_obj_t *stt_create_string(stt_ctx_t *stt, const char *val)
{
	stt_obj_t *o = stt_alloc_obj(stt);
	o->type = STT_ST_STRING;
	if(!(o->d.string.value = stt_strdup(val))) {
		perror("create_string");
		exit(STT_EXIT_MEMORY_FAIL);
	}
	return o;
}

/* create a string from n bytes */
stt_obj_t *stt_create_string_counted(stt_ctx_t *stt, const char *val, const size_t len)
{
	stt_obj_t *o = stt_alloc_obj(stt);
	o->type = STT_ST_STRING;
	o->d.string.value = (char *) stt_malloc(len + 1);
	memcpy(o->d.string.value, val, len);
	o->d.string.value[len] = 0;
	return o;
}

/* create a string (doesn't stt_strdup val) */
stt_obj_t *stt_create_string_allocd(stt_ctx_t *stt, char *val)
{
	stt_obj_t *o = stt_alloc_obj(stt);
	o->type = STT_ST_STRING;
	o->d.string.value = val;
	return o;
}

/* create a blob (use pointer, don't copy data) */
stt_obj_t *stt_create_blob(stt_ctx_t *stt, unsigned char *data, size_t size)
{
	stt_obj_t *o = stt_alloc_obj(stt);
	o->type = STT_ST_BLOB;
	o->d.blob.data = data;
	o->d.blob.size = size;
	return o;
}

/* create a blob (copy data) */
stt_obj_t *stt_create_blob_copy(stt_ctx_t *stt, unsigned char *data, size_t size)
{
	unsigned char *ndata = (unsigned char *) stt_malloc(size);
	memcpy(ndata, data, size);
	return stt_create_blob(stt, ndata, size);
}

/* create a CONS */
stt_obj_t *stt_create_cons(stt_ctx_t *stt, stt_obj_t *car, stt_obj_t *cdr)
{
	stt_obj_t *o = stt_alloc_obj(stt);
	o->type = STT_ST_CONS;
	o->d.cons.car = car;
	o->d.cons.cdr = cdr;
	return o;
}

/* create an annot */
stt_obj_t *stt_create_annot(stt_ctx_t *stt, stt_obj_t *val, stt_obj_t *annot)
{
	stt_obj_t *o = stt_alloc_obj(stt);
	o->type = STT_ST_ANNOT;
	o->d.annot.val = val;
	o->d.annot.annot = annot;
	return o;
}

/* set last element of a list */
void stt_list_setlast(stt_obj_t *lst, stt_obj_t *obj)
{
	while(lst->d.cons.cdr->type == STT_ST_CONS)
		lst = lst->d.cons.cdr;
	lst->d.cons.cdr = obj;
}

/* create an error object */
stt_obj_t *stt_create_error(stt_ctx_t *stt, const int code, const char *message)
{
	stt_obj_t *o = stt_alloc_obj(stt);
	o->type = STT_ST_ERROR;
	o->d.error.code = code;
	if(message)
		o->d.error.msg = stt_strdup(message);
	o->d.error.file = stt->lastcall_file;
	o->d.error.line = stt->lastcall_line;
	return o;
}

/* create an error object (with ->obj set) */
stt_obj_t *stt_create_error_with_obj(stt_ctx_t *stt, const int code, const char *message, stt_obj_t *obj)
{
	stt_obj_t *o = stt_create_error(stt, code, message);
	o->d.error.obj = obj;
	return o;
}

/* create a builtin object, with user data */
stt_obj_t *stt_create_builtin_with_data(stt_ctx_t *stt, stt_builtin_handler handler, void *data)
{
	stt_obj_t *o = stt_alloc_obj(stt);
	o->type = STT_ST_BUILTIN;
	o->d.builtin.handler = handler;
	o->d.builtin.data = data;
	return o;
}

/* create a builtin object */
stt_obj_t *stt_create_builtin(stt_ctx_t *stt, stt_builtin_handler handler)
{
	stt_obj_t *o = stt_alloc_obj(stt);
	o->type = STT_ST_BUILTIN;
	o->d.builtin.handler = handler;
	return o;
}

/* create an EOF object */
stt_obj_t *stt_create_eof(stt_ctx_t *stt)
{
	stt_obj_t *o;

	if(!cached_eof) {
		o = (stt_obj_t *) stt_calloc(1, sizeof(stt_obj_t));
		o->file = stt->file;
		o->line = stt->parsed_line;
		o->type = STT_ST_EOF;
		cached_eof = o;
		stt_mark_miscfree(stt, o);
		return o;
	} else return cached_eof;
}

/* create a T object */
stt_obj_t *stt_create_t(stt_ctx_t *stt)
{
	stt_obj_t *o;
	if(!cached_t) {
		o = (stt_obj_t *) stt_calloc(1, sizeof(stt_obj_t));
		o->file = stt->file;
		o->line = stt->parsed_line;
		o->type = STT_ST_T;
		cached_t = o;
		stt_mark_miscfree(stt, o);
		return o;
	} else return cached_t;
}

/* create a STT_NIL object */
stt_obj_t *stt_create_nil(stt_ctx_t *stt)
{
	stt_obj_t *o;
	if(!cached_nil) {
		o = (stt_obj_t *) stt_calloc(1, sizeof(stt_obj_t));
		o->file = stt->file;
		o->line = stt->parsed_line;
		o->type = STT_ST_NIL;
		cached_nil = o;
		stt_mark_miscfree(stt, o);
		return o;
	} else return cached_nil;
}

stt_obj_t *stt_create_new_nil(stt_ctx_t *stt)
{
	stt_obj_t *o = stt_alloc_obj(stt);
	o->type = STT_ST_NIL;
	return o;
}

stt_obj_t *stt_create_new_t(stt_ctx_t *stt)
{
	stt_obj_t *o = stt_alloc_obj(stt);
	o->type = STT_ST_T;
	return o;}

/* create a function */
stt_obj_t *stt_create_func(stt_ctx_t *stt, stt_obj_t *prm, stt_obj_t *exp)
{
	stt_obj_t *o = stt_alloc_obj(stt);
	o->type = STT_ST_FUNC;
	o->d.func.parm = prm;
	o->d.func.exp = exp;
	return o;
}

/* create a call deferral */
stt_obj_t *stt_create_defer(stt_ctx_t *stt, stt_obj_t *func, stt_obj_t *args, stt_obj_t *obj)
{
	stt_obj_t *o = stt_alloc_obj(stt);
	o->type = STT_ST_DEFER;
	o->d.defer.func = func;
	o->d.defer.args = args;
	o->d.defer.obj = obj;
	return o;
}

/* create an array */
stt_obj_t *stt_create_array(stt_ctx_t *stt, const int dim, const long *size)
{
	int i, a = 1;
	stt_obj_t *o = stt_alloc_obj(stt);
	o->type = STT_ST_ARRAY;
	o->d.array.dim = dim;

	for(i=0;i<dim; i++)
		a *= (o->d.array.size[i] = size[i]);
	o->d.array.data = (stt_obj_t **) malloc(a * sizeof(stt_obj_t *));
	if(!o->d.array.data)
		return stt_create_error(stt, STT_SE_MEM, "Insufficient memory for array allocation");
	for(i=0; i<a; i++)
		o->d.array.data[i] = stt_create_nil(stt);
	return o;
}

/* create a dictionary */
stt_obj_t *stt_create_dict(stt_ctx_t *stt)
{
	stt_obj_t *o = stt_alloc_obj(stt);
	o->type = STT_ST_DICT;
	o->d.dict.size = stt->default_dict_size;
	o->d.dict.mask = o->d.dict.size - 1;
	o->d.dict.threshold = (o->d.dict.size * stt->dict_threshold) >> 8;
	o->d.dict.data = (stt_dictentry_t *) stt_calloc(o->d.dict.size, sizeof(stt_dictentry_t));
	return o;
}

/* create a stream */
stt_obj_t *stt_create_stream(stt_ctx_t *stt, const int type)
{
	stt_obj_t *o = stt_alloc_obj(stt);
	o->type = STT_ST_STREAM;
	o->d.stream.type = type;
	o->d.stream.unread = 256;
	o->d.stream.line = 1;
	return o;
}

/* create encapsulated stt_varctx_t */
stt_obj_t *stt_create_obj_varctx(stt_ctx_t *stt, stt_varctx_t *ctx)
{
	stt_obj_t *o = stt_alloc_obj(stt);
	o->type = STT_ST_VARCTX;
	o->d.varctx.ctx = ctx;
	return o;
}

/* create cookie */
stt_obj_t *stt_create_cookie(stt_ctx_t *stt, const int type, void *ptr)
{
	stt_obj_t *o = stt_alloc_obj(stt);
	o->type = STT_ST_COOKIE;
	o->d.cookie.type = type;
	o->d.cookie.ptr = ptr;
	return o;
}

/* create class */
stt_obj_t *stt_create_class(stt_ctx_t *stt)
{
	stt_obj_t *o = stt_alloc_obj(stt);
	o->type = STT_ST_CLASS;
	o->d.clos_class = (stt_class_t *) stt_calloc(1, sizeof(stt_class_t));
	o->d.clos_class->initarg = stt_varctx_create(stt, NULL, CLASS_NAMES_SIZE);
	o->d.clos_class->slotname = stt_varctx_create(stt, NULL, CLASS_NAMES_SIZE);
	o->d.clos_class->parents = stt_create_nil(stt);
	o->d.clos_class->children = stt_create_nil(stt);
	o->d.clos_class->slots = stt_create_nil(stt);
	o->d.clos_class->prec = stt_create_nil(stt);
	o->d.clos_class->ancestry = stt_create_nil(stt);

	return o;
}

/* create slot */
stt_obj_t *stt_create_slot(stt_ctx_t *stt)
{
	stt_obj_t *o = stt_alloc_obj(stt);
	o->type = STT_ST_SLOT;
	o->d.clos_slot.initform = stt_create_nil(stt);
	return o;
}

/* create object */
stt_obj_t *stt_create_object(stt_ctx_t *stt, stt_obj_t *type)
{
	stt_obj_t *o = stt_alloc_obj(stt);
	o->type = STT_ST_OBJECT;
	o->d.clos_object.type = type;
	o->d.clos_object.data_size = type->d.clos_class->instdata;
	o->d.clos_object.data = (stt_obj_t **) stt_calloc(o->d.clos_object.data_size, sizeof(stt_obj_t *));
	return o;
}

/* create generic */
stt_obj_t *stt_create_generic(stt_ctx_t *stt)
{
	stt_obj_t *o = stt_alloc_obj(stt);
	o->type = STT_ST_GENERIC;
	o->d.clos_generic.dispatch = (stt_dispatch_entry_t *) stt_calloc(DEFAULT_DISPATCH_SIZE, sizeof(stt_dispatch_entry_t));
	o->d.clos_generic.dispatch_size = DEFAULT_DISPATCH_SIZE;
	o->d.clos_generic.dispatch_fill = 0;
	o->d.clos_generic.dispatch_limit = (DEFAULT_DISPATCH_SIZE * DISPATCH_THRESHOLD) >> 8;
	o->d.clos_generic.m_primary = stt_create_nil(stt);
	o->d.clos_generic.m_before = stt_create_nil(stt);
	o->d.clos_generic.m_after = stt_create_nil(stt);
	o->d.clos_generic.m_around = stt_create_nil(stt);
	return o;
}

/* create method */
stt_obj_t *stt_create_method(stt_ctx_t *stt)
{
	stt_obj_t *o = stt_alloc_obj(stt);
	o->type = STT_ST_METHOD;
	return o;
}

/* create dispatch state */
stt_obj_t *stt_create_dispatch(stt_ctx_t *stt)
{
	stt_obj_t *o = stt_alloc_obj(stt);
	o->type = STT_ST_DISPATCH;
	o->d.clos_dispatch.argeval = (stt_obj_t **) stt_calloc(FUNC_MAXPARMS + 1, sizeof(stt_obj_t *));
	return o;
}

/* create a varref */
stt_obj_t *stt_create_varref(stt_ctx_t *stt, ptrdiff_t offset)
{
	stt_obj_t *o = stt_alloc_obj(stt);
	o->type = STT_ST_VARREF;
	o->d.varref.offset = offset;
	return o;
}

/*********************************** STREAMS ***/

/* print array contents */
static stt_obj_t * printarray(stt_varctx_t *ctx, const stt_obj_t *ar, stt_obj_t **data, int dim)
{
	long ds, i;
	stt_obj_t *o;
	ds = 1;
	for(i=0;i<dim;i++)
		ds *= ar->d.array.size[i];
	o = stt_stream_stdout_write(ctx, '<');
	if(o->type == STT_ST_ERROR) return o;

	if(dim < ar->d.array.dim-1) {
		for(i=0;i<ar->d.array.size[dim];i++) {
			if(i) {
				o = stt_stream_stdout_write(ctx, ' ');
				if(o->type == STT_ST_ERROR) return o;
			}
			o = printarray(ctx, ar, data, dim+1);
			if(o->type == STT_ST_ERROR) return o;
			data += ds;
		}
	} else {
		for(i=0;i<ar->d.array.size[dim];i++) {
			if(i) {
				o = stt_stream_stdout_write(ctx, ' ');
				if(o->type == STT_ST_ERROR) return o;
			}
			o = stt_printobj(ctx, *data);
			if(o->type == STT_ST_ERROR) return o;
			data += ds;
		}
	}

	o = stt_stream_stdout_write(ctx, '>');
	if(o->type == STT_ST_ERROR) return o;

	return stt_create_nil(ctx->stt);
}

/* write a character to a stream */
stt_obj_t *stt_stream_write_char(stt_varctx_t *ctx, stt_obj_t *stream, int ch)
{
	if(stream->type != STT_ST_STREAM)
		return stt_create_error(ctx->stt, STT_SE_TYPE, "Type mismatch in stream_write_char");

	if(ch == '\n') stream->d.stream.line ++;

	switch(stream->d.stream.type) {
		case STT_STREAM_FP:
			if(!stream->d.stream.s.fp)
				return stt_create_error(ctx->stt, STT_SE_IO, "I/O error (stream closed)");
			if(fputc(ch, stream->d.stream.s.fp) == EOF)
				return stt_create_error(ctx->stt, STT_SE_IO, "I/O error during stream_write_char");
			break;
		case STT_STREAM_FUNC:
			return stt_funcexec(ctx, NULL, stream->d.stream.s.func, stt_create_cons(ctx->stt, stt_create_char(ctx->stt, ch), stt_create_nil(ctx->stt)), 0, 0);
		case STT_STREAM_C:
			if(!stream->d.stream.s.c.writer)
				return stt_create_error(ctx->stt, STT_SE_IO, "External stream not writable");
			if(stream->d.stream.s.c.writer(ch, &stream->d.stream.s.c.data, stream->d.stream.s.c.type))
				return stt_create_error_with_obj(ctx->stt, STT_SE_IO, "I/O error in external writer function", stt_create_string(ctx->stt, strerror(errno)));
			break;
		default:
			;
	}

	return stt_create_nil(ctx->stt);
}

/* read buffer from instream */
stt_obj_t *stt_stream_readbuf(stt_varctx_t *ctx, unsigned char *buf, size_t bytes, size_t *readlen)
{
	stt_obj_t *stream = stt_get_input_stream(ctx);
	if(stream->type == STT_ST_ERROR) return stream;

	if(stream->d.stream.type == STT_STREAM_FP) {
		size_t r = fread(buf, 1, bytes, stream->d.stream.s.fp);
		*readlen = r;
		return STT_NIL(ctx->stt);
	} else if(stream->d.stream.type == STT_STREAM_C && stream->d.stream.s.c.readmore) {
		size_t r = stream->d.stream.s.c.readmore(&stream->d.stream.s.c.data, stream->d.stream.s.c.type, buf, bytes);
		if(r == -1)
			return stt_create_error(ctx->stt, STT_SE_IO, "stream_readbuf: I/O error in readmore");
		*readlen = r;
		return STT_NIL(ctx->stt);
	} else {
		size_t l = 0;
		for(;;) {
			stt_obj_t *o = stt_stream_read_char(ctx, stream);

			if(o->type == STT_ST_ERROR) return o;

			if(o->type == STT_ST_EOF) {
				*readlen = l;
				return o;
			}

			buf[l++] = o->d.ch.ch;
			if(l == bytes) {
				*readlen = l;
				return STT_NIL(ctx->stt);
			}
		}
	}
}

/* dump buffer to outstream */
stt_obj_t *stt_stream_dumpbuf(stt_varctx_t *ctx, unsigned char *buf, size_t len)
{
	size_t i;
	stt_obj_t *str, *o = NULL;

	if(len == 0) return STT_NIL(ctx->stt);

	str = stt_get_output_stream(ctx);
	if(str->type == STT_ST_ERROR) return str;

	if(str->d.stream.type == STT_STREAM_FP) {
		if(fwrite(buf, len, 1, str->d.stream.s.fp) == 0)
			return stt_create_error(ctx->stt, STT_SE_IO, "I/O error in stream_dumpbuf FP writing");
		else
			return STT_NIL(ctx->stt);
	} else if(str->d.stream.type == STT_STREAM_C && str->d.stream.s.c.writemore) {
		size_t r = str->d.stream.s.c.writemore(&str->d.stream.s.c.data, str->d.stream.s.c.type, buf, len);
		if(r == -1) return stt_create_error(ctx->stt, STT_SE_IO, "I/O error in stream_dumpbuf C writemore");
		return stt_create_nil(ctx->stt);
	}

	for(i=0; i<len; i++) {
		o = stt_stream_write_char(ctx, str, buf[i]);
		if(o->type == STT_ST_ERROR) return o;
	}
	return o ? o : stt_create_nil(ctx->stt);
}

/* printf to the standard output stream */
stt_obj_t *stt_stream_printf(stt_varctx_t *ctx, const char *format, ...)
{
	va_list ap;
	size_t len;
	stt_obj_t *str, *o = NULL;

	str = stt_get_output_stream(ctx);
	if(str->type == STT_ST_ERROR) return str;

	va_start(ap, format);
	if (str->d.stream.type == STT_STREAM_FP) {
		len = vfprintf(str->d.stream.s.fp, format, ap);
	}
	else if(str->d.stream.type == STT_STREAM_C && str->d.stream.s.c.writemore) {
		len = str->d.stream.s.c.vprintf(&str->d.stream.s.c.data, str->d.stream.s.c.type, format, ap);
	}
	else {
		char buf[2048], *dst;
		int need2free;
		size_t r, i;

		dst = stt_asvprintf(buf, sizeof(buf), &need2free, &r, format, ap);

		for(i=0; i<r; i++) {
			o = stt_stream_write_char(ctx, str, dst[i]);
			if(o->type == STT_ST_ERROR) {
				va_end(ap);
				if (need2free)
					free(dst);
				return o;
			}
		}
		va_end(ap);
		if (need2free)
			free(dst);
		return o ? o : stt_create_nil(ctx->stt);
	}
	va_end(ap);

	if (len < 0)
		return stt_create_error(ctx->stt, STT_SE_IO, "I/O error in stream_printf: stream type does not implement printf");

	return stt_create_nil(ctx->stt);
}

/* unread a character into a stream */
stt_obj_t *stt_stream_unread_char(stt_ctx_t *stt, stt_obj_t *stream, const unsigned char ch)
{
	if(stream->type != STT_ST_STREAM)
		return stt_create_error(stt, STT_SE_TYPE, "Type mismatch in stream_unread_char");
	if(stream->d.stream.unread != 256)
		return stt_create_error(stt, STT_SE_IO, "Can't unread two characters without reading one in the middle");

	stream->d.stream.unread = ch;
	return stream;
}

/* read a character form a stream */
stt_obj_t *stt_stream_read_char(stt_varctx_t *ctx, stt_obj_t *stream)
{
	stt_obj_t *obj;
	int ch;

	if(stream->type != STT_ST_STREAM)
		return stt_create_error(ctx->stt, STT_SE_TYPE, "Type mismatch in stream_read_char");

	if(stream->d.stream.unread != 256) {
		/* there is an unread character, let's give it back. */
		obj = stt_create_char(ctx->stt, stream->d.stream.unread);
		stream->d.stream.unread = 256;
		return obj;
	}

	switch(stream->d.stream.type) {
		case STT_STREAM_FP:
			if(!stream->d.stream.s.fp)
				return stt_create_error(ctx->stt, STT_SE_IO, "I/O error (stream closed)");
			ch = fgetc(stream->d.stream.s.fp);
			if(ch == '\n') ctx->stt->parsed_line = stream->d.stream.line ++;
			if(ch == EOF)
				return stt_create_eof(ctx->stt);
			return stt_create_char(ctx->stt, ch);
		case STT_STREAM_FUNC:
			obj = stt_funcexec(ctx, NULL, stream->d.stream.s.func, stt_create_nil(ctx->stt), 0, 0);
			switch(obj->type) {
				case STT_ST_STRING:
					if(obj->d.string.value[0] == '\n') ctx->stt->parsed_line = stream->d.stream.line ++;
					return stt_create_char(ctx->stt, obj->d.string.value[0]);
				case STT_ST_CHAR:
					if(obj->d.ch.ch == '\n') ctx->stt->parsed_line = stream->d.stream.line ++;
				case STT_ST_ERROR:
				case STT_ST_EOF:
					return obj;
				default:
					return stt_create_error(ctx->stt, STT_SE_TYPE, "Type mismatch in stream_read_char");
			}
		case STT_STREAM_C:
			if(!stream->d.stream.s.c.reader)
				return stt_create_error(ctx->stt, STT_SE_IO, "External stream not readable");
			ch = stream->d.stream.s.c.reader(&stream->d.stream.s.c.data, stream->d.stream.s.c.type);
			if(ch == '\n') ctx->stt->parsed_line = stream->d.stream.line ++;
			if(ch == EOF)
				return stt_create_eof(ctx->stt);
			return stt_create_char(ctx->stt, ch);
		default:
			return stt_create_error(ctx->stt, STT_SE_TYPE, "Invalid stream type in stream_read_char");
	}

	return stt_create_nil(ctx->stt);
}

/* returns the value of *output-stream* */
stt_obj_t *stt_get_output_stream(stt_varctx_t *ctx)
{
	static unsigned long output_stream_sdbm = 0;
	static unsigned long output_stream_id = 0;
	stt_var_t *v;

	if(!output_stream_id) {
		output_stream_sdbm = stt_sdbm("*OUTPUT-STREAM*");
		output_stream_id = stt_get_symbolid(ctx->stt, "*OUTPUT-STREAM*", output_stream_sdbm);
	}

	v = stt_varctx_get_s(ctx, output_stream_sdbm, output_stream_id);
	if(!v)
		return stt_create_error(ctx->stt, STT_SE_WTF, "*OUTPUT-STREAM* undefined");
	if(v->value->type != STT_ST_STREAM)
		return stt_create_error(ctx->stt, STT_SE_TYPE, "*OUTPUT-STREAM* not a stream");

	return v->value;
}

/* returns the value of *input-stream* */
stt_obj_t *stt_get_input_stream(stt_varctx_t *ctx)
{
	static unsigned long input_stream_sdbm = 0;
	static unsigned long input_stream_id = 0;
	stt_var_t *v;

	if(!input_stream_id) {
		input_stream_sdbm = stt_sdbm("*INPUT-STREAM*");
		input_stream_id = stt_get_symbolid(ctx->stt, "*INPUT-STREAM*", input_stream_sdbm);
	}

	v = stt_varctx_get_s(ctx, input_stream_sdbm, input_stream_id);
	if(!v)
		return stt_create_error(ctx->stt, STT_SE_WTF, "*INPUT-STREAM* undefined");
	if(v->value->type != STT_ST_STREAM)
		return stt_create_error(ctx->stt, STT_SE_TYPE, "*INPUT-STREAM* not a stream");

	return v->value;
}

/* read a character from *input-stream* */
stt_obj_t *stt_stream_stdin_read(stt_varctx_t *ctx)
{
	stt_obj_t *str = stt_get_input_stream(ctx);
	if(str->type == STT_ST_ERROR) return str;
	return stt_stream_read_char(ctx, str);
}

/* unread a character into *input-stream* */
stt_obj_t *stt_stream_stdin_unread(stt_varctx_t *ctx, const unsigned char ch)
{
	stt_obj_t *str = stt_get_input_stream(ctx);
	if(str->type == STT_ST_ERROR) return str;
	return stt_stream_unread_char(ctx->stt, str, ch);
}

/* write a character to *output-stream* */
stt_obj_t *stt_stream_stdout_write(stt_varctx_t *ctx, int ch)
{
	stt_obj_t *str = stt_get_output_stream(ctx);
	if(str->type == STT_ST_ERROR) return str;
	return stt_stream_write_char(ctx, str, ch);
}

/* print object to a FP (for debugging) */
stt_obj_t * stt_fp_printobj(FILE *fp, stt_varctx_t *ctx, const stt_obj_t *obj)
{
	stt_var_t *v = stt_varctx_get(ctx, "*OUTPUT-STREAM*");
	stt_obj_t *str, *val, *ret;

	if(!v) return stt_create_error(ctx->stt, STT_SE_WTF, "*OUTPUT-STREAM* undefined");
	val = v->value;

	str = stt_create_stream(ctx->stt, STT_STREAM_FP);
	str->d.stream.s.fp = fp;
	v->value = str;

	ret = stt_printobj(ctx, obj);

	v->value = val;
	return ret;
}

/* print description/value of an object */
stt_obj_t * stt_printobj(stt_varctx_t *ctx, const stt_obj_t *obj)
{
	stt_obj_t *o, *oi;
	char *c;
	int i;
	char buf[32];

	if(!obj) {
		o = stt_stream_printf(ctx, "NULL");
		return o;
	}

	if(obj->type == STT_ST_ANNOT)
		obj = obj->d.annot.val;

	switch(obj->type) {
		/* string */
		case STT_ST_STRING:
			o = stt_stream_stdout_write(ctx, '"');
			if(o->type == STT_ST_ERROR) return o;
			for(c = obj->d.string.value; *c; c++) {
				if(*c == '\\' || *c == '\"') {
					o = stt_stream_stdout_write(ctx, '\\');
					if(o->type == STT_ST_ERROR) return o;
				}
				o = stt_stream_stdout_write(ctx, *c);
				if(o->type == STT_ST_ERROR) return o;
			}
			o = stt_stream_stdout_write(ctx, '"');
			return o;

		/* double */
		case STT_ST_NUM:
			sprintf(buf, "%.12f", obj->d.num.value);
			c = buf + strlen(buf) - 1;
			for(;;) {
				if(*c == '.') {
					*++c = '0';
					break;
				} else if(*c == '0') {
					*c-- = 0;
				} else break;
			}
			return stt_stream_printf(ctx, buf);

		/* integer */
		case STT_ST_INT:
			return stt_stream_printf(ctx, "%ld", obj->d.integer.value);

		/* character */
		case STT_ST_CHAR:
			return stt_stream_printf(ctx, "#\\%c", obj->d.ch.ch);

		/* symbol */
		case STT_ST_SYMBOL:
			if(obj->d.symbol.colon) {
				o = stt_stream_stdout_write(ctx, ':');
				if(o->type == STT_ST_ERROR) return o;
			}
			return stt_stream_printf(ctx, "%s", obj->d.symbol.name);

		/* array */
		case STT_ST_ARRAY:
			o = stt_stream_printf(ctx, "#A");
			if(o->type == STT_ST_ERROR) return o;
			return printarray(ctx, obj, obj->d.array.data, 0);

		/* dictionary */
		case STT_ST_DICT:
			o = stt_stream_printf(ctx, "#D<");
			if(o->type == STT_ST_ERROR) return o;
			{
				unsigned long i;
				for(i=0; i<obj->d.dict.size; i++) {
					if(obj->d.dict.data[i].name) {
						o = stt_stream_stdout_write(ctx, ' ');
						if(o->type == STT_ST_ERROR) return o;
						o = stt_printobj(ctx, obj->d.dict.data[i].name);
						if(o->type == STT_ST_ERROR) return o;
						o = stt_stream_stdout_write(ctx, ':');
						if(o->type == STT_ST_ERROR) return o;
						o = stt_printobj(ctx, obj->d.dict.data[i].value);
						if(o->type == STT_ST_ERROR) return o;
					}
				}
			}
			return stt_stream_printf(ctx, " >");

		/* STT_NIL */
		case STT_ST_NIL:
			return stt_stream_printf(ctx, "()");

		/* T */
		case STT_ST_T:
			return stt_stream_printf(ctx, "T");

		/* BLOB */
		case STT_ST_BLOB:
			return stt_stream_printf(ctx, "#BLOB<%ld>", (long)obj->d.blob.size);

		/* error */
		case STT_ST_ERROR:
			o = stt_stream_printf(ctx, "ERROR %s: %s at %s:%ld", stt_errorname(ctx->stt, obj->d.error.code), obj->d.error.msg, obj->d.error.file, obj->d.error.line);
			if(o->type == STT_ST_ERROR) return o;

			if(obj->d.error.obj) {
				o = stt_stream_printf(ctx, " [");
				if(o->type == STT_ST_ERROR) return o;
				o = stt_printobj(ctx, obj->d.error.obj);
				if(o->type == STT_ST_ERROR) return o;
				o = stt_stream_printf(ctx, "]");
				if(o->type == STT_ST_ERROR) return o;
			}

			o = stt_stream_printf(ctx, "\n");
			if(o->type == STT_ST_ERROR) return o;

			/* print error trace */
			for(oi = obj->d.error.trace; oi && oi->type == STT_ST_CONS; oi=oi->d.cons.cdr) {
				if(oi->d.cons.car->type == STT_ST_ERROR) {
					o = stt_stream_printf(ctx, "\n  ");
					if(o->type == STT_ST_ERROR) return o;

					if(oi->d.cons.car->d.error.obj) {
						stt_obj_t *src = oi->d.cons.car->d.error.obj;

						o = stt_printobj(ctx, src);
						if(o->type == STT_ST_ERROR) return o;
					}

					o = stt_stream_printf(ctx, "\n -------------------------* %s:%ld *---------",
							oi->d.cons.car->d.error.file,
							oi->d.cons.car->d.error.line);
					if(o->type == STT_ST_ERROR) return o;
				}
			}
			break;

		/* builtin function */
		case STT_ST_BUILTIN:
			if(obj->d.builtin.symbol)
				return stt_stream_printf(ctx, "%s", obj->d.builtin.symbol);
			else
				return stt_stream_printf(ctx, "#b:%p", obj->d.builtin.handler);

		/* stream */
		case STT_ST_STREAM:
			return stt_stream_printf(ctx, "#STREAM");

		/* function */
		case STT_ST_FUNC:
			if(obj->d.func.symbol)
				o = stt_stream_printf(ctx, "#%c:%s", obj->d.func.macro ? 'm' : 'l', obj->d.func.symbol);
			else
				o = stt_stream_printf(ctx, obj->d.func.macro ? "macro" : "fn");
			if(o->type == STT_ST_ERROR) return o;
			if(obj->d.func.parm->type == STT_ST_CONS)
				o = stt_printobj(ctx, obj->d.func.parm);
			else
				o = stt_stream_printf(ctx, "()");
			if(o->type == STT_ST_ERROR) return o;
			if(obj->d.func.docstr) {
				o = stt_stream_stdout_write(ctx, ' ');
				if(o->type == STT_ST_ERROR) return o;
				o = stt_printobj(ctx, obj->d.func.docstr);
			}
			return o;

		/* deferred execution (doesn't happen) */
		case STT_ST_DEFER:
			return stt_stream_printf(ctx, "#DEFER");

		/* encapsulated stt_varctx_t */
		case STT_ST_VARCTX:
			return stt_stream_printf(ctx, "#VARCTX:%p", obj->d.varctx.ctx);

		/* cookie */
		case STT_ST_COOKIE:
			return stt_stream_printf(ctx, "#COOKIE:%d", obj->d.cookie.type);

		/**** CLOS ****/

		/* class */
		case STT_ST_CLASS:
			return stt_stream_printf(ctx, "#CLASS");

		/* slot */
		case STT_ST_SLOT:
			return stt_stream_printf(ctx, "#SLOT");

		/* instance */
		case STT_ST_OBJECT:
			return stt_stream_printf(ctx, "#OBJECT<%s>", obj->d.clos_object.type->d.clos_class->name->d.symbol.name);

		/* method */
		case STT_ST_METHOD:
			o = stt_stream_printf(ctx, "\n  #METHOD<%p>[code=%p, args={", obj, obj->d.clos_method.code);
			if(o->type == STT_ST_ERROR) return o;

			for(i=0; i<obj->d.clos_method.argn; i++) {
				o = stt_stream_stdout_write(ctx, ' ');
				if(o->type == STT_ST_ERROR) return o;

				o = stt_printobj(ctx, obj->d.clos_method.args[i].name);
				if(o->type == STT_ST_ERROR) return o;
				o = stt_stream_printf(ctx, ",%d,", obj->d.clos_method.args[i].type);
				if(o->type == STT_ST_ERROR) return o;
				o = stt_printobj(ctx, obj->d.clos_method.args[i].value);
				if(o->type == STT_ST_ERROR) return o;
			}

			o = stt_stream_printf(ctx, " }; miscargs=");
			if(o->type == STT_ST_ERROR) return o;

			o = stt_printobj(ctx, obj->d.clos_method.miscargs);
			if(o->type == STT_ST_ERROR) return o;

			return stt_stream_printf(ctx, "]");

		/* dispatch info */
		case STT_ST_DISPATCH:
			return stt_stream_printf(ctx, "#DISPATCH");

		/* generic */
		case STT_ST_GENERIC:
			o = stt_stream_printf(ctx, "#GENERIC[\nprimary=");
			if(o->type == STT_ST_ERROR) return o;

			o = stt_printobj(ctx, obj->d.clos_generic.m_primary);
			if(o->type == STT_ST_ERROR) return o;

			o = stt_stream_printf(ctx, "\nbefore=");
			if(o->type == STT_ST_ERROR) return o;

			o = stt_printobj(ctx, obj->d.clos_generic.m_before);
			if(o->type == STT_ST_ERROR) return o;

			o = stt_stream_printf(ctx, "\nafter=");
			if(o->type == STT_ST_ERROR) return o;

			o = stt_printobj(ctx, obj->d.clos_generic.m_after);
			if(o->type == STT_ST_ERROR) return o;

			o = stt_stream_printf(ctx, "\naround=");
			if(o->type == STT_ST_ERROR) return o;

			o = stt_printobj(ctx, obj->d.clos_generic.m_around);
			if(o->type == STT_ST_ERROR) return o;

			return stt_stream_printf(ctx, "]");

		/**** /CLOS ****/

		/* EOF */
		case STT_ST_EOF:
			return stt_stream_printf(ctx, "EOF");

		/* cons */
		case STT_ST_CONS:
			/* try to print 'A instead of (b:something A) */
			if(obj->d.cons.car->type == STT_ST_BUILTIN &&
					obj->d.cons.car->d.builtin.printprefix != 0 &&
					obj->d.cons.cdr->type == STT_ST_CONS &&
					obj->d.cons.cdr->d.cons.cdr->type == STT_ST_NIL) {
				o = stt_stream_stdout_write(ctx, obj->d.cons.car->d.builtin.printprefix);
				if(o->type == STT_ST_ERROR) return o;
				return stt_printobj(ctx, obj->d.cons.cdr->d.cons.car);
			} else {
				/* just a standard cons.. */
				o = stt_stream_stdout_write(ctx, '(');
				if(o->type == STT_ST_ERROR) return o;
				o = stt_printobj(ctx, obj->d.cons.car);
				if(o->type == STT_ST_ERROR) return o;

				if(obj->d.cons.cdr != NULL && obj->d.cons.cdr->type != STT_ST_NIL) {
					if(obj->d.cons.cdr->type != STT_ST_CONS) {
						o = stt_stream_printf(ctx, " . ");
						if(o->type == STT_ST_ERROR) return o;
						o = stt_printobj(ctx, obj->d.cons.cdr);
						if(o->type == STT_ST_ERROR) return o;
					} else {
						for(oi = obj->d.cons.cdr; oi && oi->type != STT_ST_NIL;) {
							o = stt_stream_stdout_write(ctx, ' ');
							if(o->type == STT_ST_ERROR) return o;
							o = stt_printobj(ctx, oi->d.cons.car);
							if(o->type == STT_ST_ERROR) return o;

							oi = oi->d.cons.cdr;
							if(oi && oi->type != STT_ST_CONS && oi->type != STT_ST_NIL) {
								o = stt_stream_printf(ctx, " . ");
								if(o->type == STT_ST_ERROR) return o;
								o = stt_printobj(ctx, oi);
								if(o->type == STT_ST_ERROR) return o;
								break;
							}
						}
					}
				}

				return stt_stream_stdout_write(ctx, ')');
			}
			break;

		/* dunno (doesn't happen) */
		default:
			return stt_stream_printf(ctx, "?");
	}

	return stt_create_nil(ctx->stt);
}

/******************************************** UTILS ***/

/* returns new type id, registers random seed */
int stt_gettypeid(stt_ctx_t *stt) {
	int id;

	if(!stt->typedb) {
		stt->typedb = (stt_typeinfo_t *) stt_calloc(128, sizeof(stt_typeinfo_t));
		stt->typedb_size = 128;
	} else {
		if(stt->next_typeid == stt->typedb_size) {
			stt->typedb_size *= 2;
			stt->typedb = (stt_typeinfo_t *) stt_realloc(stt->typedb, stt->typedb_size * sizeof(stt_typeinfo_t));
		}
	}

	stt->typedb[stt->next_typeid].typeseed = stt_hash((unsigned char *) &stt->next_typeid, sizeof(int), 0);
	stt->typedb[stt->next_typeid].cl = NULL;
	id = stt->next_typeid++;

	return id;
}

/* get type id for symbol. -1 if no such thing */
int stt_symbol_to_typeid(stt_ctx_t *stt, stt_obj_t *sym)
{
	stt_var_t *var;
	assert(sym->type == STT_ST_SYMBOL);

	/* check if a class exists by that name */
	var = stt_varctx_get_s(stt->class_hash, sym->d.symbol.sdbm, sym->d.symbol.id);
	if(var) return var->value->d.clos_class->typeid; /* yep, return type id */

	/* check for atomic types
	 * typeid should be equal to the value of the type constant in the enum */

	if(!strcmp(sym->d.symbol.name, "NIL"))
		return STT_ST_NIL;
	if(!strcmp(sym->d.symbol.name, "CONS"))
		return STT_ST_CONS;
	if(!strcmp(sym->d.symbol.name, "INT"))
		return STT_ST_INT;
	if(!strcmp(sym->d.symbol.name, "NUM"))
		return STT_ST_NUM;
	if(!strcmp(sym->d.symbol.name, "CHAR"))
		return STT_ST_CHAR;
	if(!strcmp(sym->d.symbol.name, "BLOB"))
		return STT_ST_BLOB;
	if(!strcmp(sym->d.symbol.name, "T"))
		return STT_ST_T;
	if(!strcmp(sym->d.symbol.name, "STRING"))
		return STT_ST_STRING;
	if(!strcmp(sym->d.symbol.name, "SYMBOL"))
		return STT_ST_SYMBOL;
	if(!strcmp(sym->d.symbol.name, "COOKIE"))
		return STT_ST_COOKIE;
	if(!strcmp(sym->d.symbol.name, "EOF"))
		return STT_ST_EOF;
	if(!strcmp(sym->d.symbol.name, "BUILTIN"))
		return STT_ST_BUILTIN;
	if(!strcmp(sym->d.symbol.name, "FUNC"))
		return STT_ST_FUNC;
	if(!strcmp(sym->d.symbol.name, "ERROR"))
		return STT_ST_ERROR;
	if(!strcmp(sym->d.symbol.name, "DEFER"))
		return STT_ST_DEFER;
	if(!strcmp(sym->d.symbol.name, "ARRAY"))
		return STT_ST_ARRAY;
	if(!strcmp(sym->d.symbol.name, "VARCTX"))
		return STT_ST_VARCTX;
	if(!strcmp(sym->d.symbol.name, "DICT"))
		return STT_ST_DICT;
	if(!strcmp(sym->d.symbol.name, "STREAM"))
		return STT_ST_STREAM;
	if(!strcmp(sym->d.symbol.name, "CLASS"))
		return STT_ST_CLASS;
	if(!strcmp(sym->d.symbol.name, "SLOT"))
		return STT_ST_SLOT;
	if(!strcmp(sym->d.symbol.name, "OBJECT"))
		return STT_ST_OBJECT;
	if(!strcmp(sym->d.symbol.name, "GENERIC"))
		return STT_ST_GENERIC;
	if(!strcmp(sym->d.symbol.name, "METHOD"))
		return STT_ST_METHOD;
	if(!strcmp(sym->d.symbol.name, "ANNOT"))
		return STT_ST_ANNOT;

	return -1;
}

/* return type id of object */
int stt_object_typeid(stt_obj_t *obj)
{
	/* if it's a CLOS object, return type id of class */
	if(obj->type == STT_ST_OBJECT)
		return obj->d.clos_object.type->d.clos_class->typeid;

	/* otherwise, return stutter type */
	return obj->type;
}

/* get hash value for an object */
unsigned long stt_hashfunc(const stt_obj_t *obj) {
	switch(obj->type) {
		case STT_ST_SYMBOL:
			return obj->d.symbol.sdbm;
		case STT_ST_STRING:
			return stt_sdbm(obj->d.string.value);
		case STT_ST_INT:
			return stt_hash((unsigned char *) &obj->d.integer.value, sizeof(obj->d.integer.value), 0);
		case STT_ST_NUM:
			return stt_hash((unsigned char *) &obj->d.num.value, sizeof(obj->d.num.value), 0);
		default:
			return (unsigned long) obj;
	}
}

/* copy object with inner data */
void stt_deepcopy(stt_ctx_t *stt, stt_obj_t *tgt, stt_obj_t *obj) {
	int i;
	long ds;

	memcpy(tgt, obj, sizeof(stt_obj_t));
/*	tgt->ref = NULL;*/

	switch(obj->type) {
		case STT_ST_STRING:
			tgt->d.string.value = stt_strdup(obj->d.string.value);
			break;
		case STT_ST_SYMBOL:
			tgt->d.symbol.name = stt_strdup(obj->d.symbol.name);
			break;
		case STT_ST_ERROR:
			tgt->d.error.msg = stt_strdup(obj->d.error.msg);
			break;
		case STT_ST_ARRAY:
			ds = 1;
			for(i=0;i<obj->d.array.dim;i++)
				ds *= obj->d.array.size[i];
			tgt->d.array.data = (stt_obj_t **) malloc(ds * sizeof(stt_obj_t *));
			if(!tgt->d.array.data)
				tgt = stt_create_error(stt, STT_SE_MEM, "Insufficient memory for deep copy");
			else {
				memcpy(tgt->d.array.data, obj->d.array.data, ds * sizeof(stt_obj_t *));
			}
			break;
		case STT_ST_DICT:
			tgt->d.dict.data = (stt_dictentry_t *) stt_calloc(obj->d.dict.size, sizeof(stt_dictentry_t));
			memcpy(tgt->d.dict.data, obj->d.dict.data, obj->d.dict.size * sizeof(stt_dictentry_t));
			break;

		case STT_ST_OBJECT:
			tgt->d.clos_object.data = (stt_obj_t **) stt_calloc(obj->d.clos_object.data_size, sizeof(stt_obj_t *));
			memcpy(tgt->d.clos_object.data, obj->d.clos_object.data, obj->d.clos_object.data_size * sizeof(stt_obj_t *));
			break;

		default:
			;
	}
}

/* remove stt_weakptr_t references */
STT_INLINE void stt_weak_unref(stt_obj_t **o)
{
	stt_weakptr_t *ptr, *n;
	if(!*o) return;
	for(ptr = (*o)->weak; ptr; ) {
		if(ptr->ptr == o) {
			n = ptr->next;
			free(ptr);
			ptr = n;
			continue;
		}

		ptr = ptr->next;
	}
}

/* free inner data of object */
void stt_freedata(stt_obj_t *obj) {
	int i;

	switch(obj->type) {
		case STT_ST_DICT:
			free(obj->d.dict.data);
			break;
		case STT_ST_ARRAY:
			free(obj->d.array.data);
			break;
		case STT_ST_CLASS:
			stt_varctx_free(obj->d.clos_class->initarg);
			stt_varctx_free(obj->d.clos_class->slotname);
			free(obj->d.clos_class->allslots);
			free(obj->d.clos_class->data);
			free(obj->d.clos_class);
			break;
		case STT_ST_OBJECT:
			free(obj->d.clos_object.data);
			break;
		case STT_ST_METHOD:
			free(obj->d.clos_method.args);
			break;
		case STT_ST_DISPATCH:
			free(obj->d.clos_dispatch.argeval);
			break;
		case STT_ST_GENERIC:
			for(i=0; i<obj->d.clos_generic.dispatch_size; i++)
				free(obj->d.clos_generic.dispatch[i].signature);

			free(obj->d.clos_generic.dispatch);
			break;
		case STT_ST_SYMBOL:
			free(obj->d.symbol.name);
			break;
		case STT_ST_STRING:
			free(obj->d.string.value);
			break;
		case STT_ST_BLOB:
			free(obj->d.blob.data);
			break;
		case STT_ST_VARCTX:
			stt_varctx_free(obj->d.varctx.ctx);
			break;
		case STT_ST_ERROR:
			if(obj->d.error.msg)
				free(obj->d.error.msg);
			break;
		case STT_ST_COOKIE:
			if(obj->d.cookie.destroy)
				obj->d.cookie.destroy(obj);
			break;
		case STT_ST_STREAM:
			if(obj->d.stream.type == STT_STREAM_C && obj->d.stream.s.c.destroy)
				obj->d.stream.s.c.destroy(obj);
			break;
		default:
			;
	}
}

/* returns true if two objects are stt_isoform(all fields are equal) */
int stt_isoform(const stt_obj_t *a, const stt_obj_t *b)
{
	size_t i;
	long s;
	if(a == NULL && b != NULL) return 0;
	if(a != NULL && b == NULL) return 0;
	if(a == NULL && b == NULL) return 1;

	if(a->type != b->type) return 0;
	switch(a->type) {
		case STT_ST_INT:
			return a->d.integer.value == b->d.integer.value;
		case STT_ST_NUM:
			return a->d.num.value == b->d.num.value;
		case STT_ST_STRING:
			return !strcmp(a->d.string.value, b->d.string.value);
		case STT_ST_CHAR:
			return a->d.ch.ch == b->d.ch.ch;
		case STT_ST_SYMBOL:
			return a->d.symbol.colon == b->d.symbol.colon && a->d.symbol.id == b->d.symbol.id;
		case STT_ST_NIL:
		case STT_ST_T:
		case STT_ST_EOF:
			return 1;
		case STT_ST_CONS:
			return stt_isoform(a->d.cons.car, b->d.cons.car) &&
				stt_isoform(a->d.cons.cdr, b->d.cons.cdr);
		case STT_ST_ARRAY:
			if(a->d.array.dim != b->d.array.dim) return 0;
			if(memcmp(a->d.array.size, b->d.array.size, a->d.array.dim * sizeof(int))) return 0;
			for(s=1,i=0;i<a->d.array.dim;i++)
				s *= a->d.array.size[i];
			for(i=0;i<s;i++)
				if(!stt_isoform(a->d.array.data[i], b->d.array.data[i])) return 0;
			return 1;
		case STT_ST_OBJECT:
			if(a->d.clos_object.type != b->d.clos_object.type) return 0;
			for(i=0; i<a->d.clos_object.data_size; i++) {
				if(!stt_isoform(a->d.clos_object.data[i], b->d.clos_object.data[i])) return 0;
			}

			return 1;
		default:
			return a == b;
	}
}

STT_BUILTIN(stt_sb_quote);

/* initialize reusable object caches */
void stt_objcache_init(stt_ctx_t *stt)
{
	long i;

#ifndef DISABLE_INTCACHE
	for(i=0; i<STT_REUSABLE_INT_LIMIT; i++) {
		stt->reusable_ints[i] = stt_alloc_ungc_obj(stt);
		stt->reusable_ints[i]->type = STT_ST_INT;
		stt->reusable_ints[i]->d.integer.value = i;

		stt_mark_miscfree(stt, (void *)stt->reusable_ints[i]);
	}
#endif

#ifndef DISABLE_CHARCACHE
	for(i=0; i<256; i++) {
		stt->reusable_chars[i] = stt_alloc_ungc_obj(stt);

		stt->reusable_chars[i]->type = STT_ST_CHAR;
		stt->reusable_chars[i]->d.ch.ch = i;

		stt_mark_miscfree(stt, (void *)stt->reusable_chars[i]);
	}
#endif

	stt->quote_builtin = stt_create_builtin(stt, stt_sb_quote);
	stt->quote_builtin->d.builtin.printprefix = '\'';
}

/* add a weak reference place to object */
void stt_weak_reference(stt_obj_t *host, stt_serial_t serial, stt_obj_t **ptr)
{
	stt_weakptr_t *wp;

/*	fprintf(stderr, "weak_reference host=%p serial=%d ptr=%p obj=%p\n", host, (int)serial, ptr, *ptr);*/

	for(wp = (*ptr)->weak; wp; wp=wp->next)
		if(wp->ptr == ptr) {
			wp->host = host;
			wp->serial = serial;
			return;
		}

	wp = (stt_weakptr_t *) stt_malloc(sizeof(stt_weakptr_t));
	wp->ptr = ptr;
	wp->host = host;
	wp->serial = serial;
	wp->next = (*ptr)->weak;
	(*ptr)->weak = wp;
}

/* check weakptrs, clean up invalid entries
 * returns number of entries removed */
int stt_weak_check(stt_obj_t *obj)
{
	stt_weakptr_t *wp, *n, *pr;

	int c = 0;

	pr = NULL;

	for(wp = obj->weak; wp; ) {
		if(wp->host->serial != wp->serial ||
				(* wp->ptr) != obj) {

			if(pr) pr->next = wp->next; else obj->weak = wp->next;

			n = wp;
			wp = wp->next;
			free(n);
			c++;
			continue;
		}

		pr = wp;
		wp = wp->next;
	}

	return c;
}


/****************************************************** CLOS ***/

/* add tree branch to the ancestry of a class */
static void build_class_ancestry(stt_ctx_t *stt, stt_obj_t *root, stt_obj_t *cl)
{
	stt_obj_t *o, *iter, *last;

	/* add self to family tree, in order */
	if(cl->d.clos_class->ancestry->type == STT_ST_NIL) {
		cl->d.clos_class->ancestry = stt_create_cons(stt, root, stt_create_nil(stt));
	} else {
		last = NULL;
		iter = cl->d.clos_class->ancestry;
		while(iter->type == STT_ST_CONS && iter->d.cons.car->d.clos_class->prec_score < root->d.clos_class->prec_score) {
			last = iter;
			iter = iter->d.cons.cdr;
		}

		/* if prec_score is equal (same class), don't add again */
		if(iter->type == STT_ST_CONS && iter->d.cons.car->d.clos_class->prec_score == root->d.clos_class->prec_score) {
			return;
		}

		if(last) {
			last->d.cons.cdr = stt_create_cons(stt, root, last->d.cons.cdr);
		} else {
			cl->d.clos_class->ancestry = stt_create_cons(stt, root, cl->d.clos_class->ancestry);
		}
	}

	/* recurse into parents */
	iter = root->d.clos_class->parents;
	while(stt_nextparm(&o, &iter)) {
		build_class_ancestry(stt, o, cl);
	}
}

/* add slots from class1 to slot assignments of class2
 * slots with the same name don't get overwritten
 * updates class2->instdata and data_size */
static void slots_merge(stt_obj_t *class1, stt_obj_t *class2)
{
	stt_obj_t *slot, *iter;
	size_t i;
	stt_slotasg_t *asg;

	iter = class1->d.clos_class->slots;
	while(stt_nextparm(&slot, &iter)) {
		/* scan class2's slot assignments for current slot name */
		for(i=0; i<class2->d.clos_class->allslotsnum; i++) {
			if(class2->d.clos_class->allslots[i].slot->d.clos_slot.name->d.symbol.id == slot->d.clos_slot.name->d.symbol.id) break;
		}

		/* no such slot found */
		if(i == class2->d.clos_class->allslotsnum) {
			class2->d.clos_class->allslots = (stt_slotasg_t *) stt_realloc(class2->d.clos_class->allslots, (class2->d.clos_class->allslotsnum + 1) * sizeof(stt_slotasg_t));

			asg = class2->d.clos_class->allslots + (class2->d.clos_class->allslotsnum ++);

			asg->slot = slot;

			if(slot->d.clos_slot.alloc == CL_ALLOC_INSTANCE) {
				asg->stor = NULL;
				asg->num = (class2->d.clos_class->instdata ++);
			} else {
				asg->stor = class2;
				asg->num = (class2->d.clos_class->data_size ++);
			}
		}
	}
}

/* calculate and update internal data of class */
stt_obj_t *stt_class_calculate(stt_ctx_t *stt, stt_obj_t *cl)
{
	stt_obj_t *o, *iter, *iter2;
	size_t i;
	stt_slotasg_t *asg;

	/* add self to parents' children */
	iter = cl->d.clos_class->parents;
	while(stt_nextparm(&o, &iter)) {
		iter2 = o->d.clos_class->children;
		if(iter2->type == STT_ST_NIL) {
			/* we are the first child */
			o->d.clos_class->children = stt_create_cons(stt, cl, stt_create_nil(stt));
		} else {
			/* append to list */
			while(iter2->d.cons.cdr->type == STT_ST_CONS)
				iter2 = iter2->d.cons.cdr;

			iter2->d.cons.cdr = stt_create_cons(stt, cl, stt_create_nil(stt));
		}
	}

	/* rebuild class precedence data */
	stt_build_precedence(stt);

	/* build sorted family tree */
	iter = cl->d.clos_class->parents;
	while(stt_nextparm(&o, &iter)) {
		build_class_ancestry(stt, o, cl);
	}

	/* calculate slot assignments */
	slots_merge(cl, cl);
	iter = cl->d.clos_class->ancestry;
	while(stt_nextparm(&o, &iter)) {
		slots_merge(o, cl);
	}

	/* fill in slotname->assignment varctxes */
	for(i=0, asg = cl->d.clos_class->allslots; i<cl->d.clos_class->allslotsnum; i++, asg++) {
		stt_varctx_set_s(cl->d.clos_class->slotname, asg->slot->d.clos_slot.name->d.symbol.sdbm, asg->slot->d.clos_slot.name->d.symbol.id, stt_create_int(stt, i));

		if(asg->slot->d.clos_slot.initarg) {
			stt_varctx_set_s(cl->d.clos_class->initarg, asg->slot->d.clos_slot.name->d.symbol.sdbm, asg->slot->d.clos_slot.name->d.symbol.id, stt_create_int(stt, i));
		}
	}

	/* allocate space for static data */
	cl->d.clos_class->data = (stt_obj_t **) stt_calloc(cl->d.clos_class->data_size, sizeof(stt_obj_t *));

	return cl;
}

/* walk a branch of class precedence */
static void build_precedence_branch(stt_ctx_t *stt, stt_obj_t *root, stt_obj_t **tail)
{
	stt_obj_t *o, *sib, *iter;
	stt_obj_t *iter2;
	int branch;

	/* add root to precedence */
	if(*tail) {
		(*tail)->d.cons.cdr = stt_create_cons(stt, root, stt_create_nil(stt));
		*tail = (*tail)->d.cons.cdr;
	} else {
		stt->class_precedence = *tail = stt_create_cons(stt, root, stt_create_nil(stt));
	}
	root->d.clos_class->prec = *tail;

	/* walk parents */
	iter = root->d.clos_class->parents;
	while(stt_nextparm(&o, &iter)) {

		/* check whether there are unexplored siblings */
		iter2 = o->d.clos_class->children;
		branch = 1;
		while(stt_nextparm(&sib, &iter2)) {
			if(!sib->d.clos_class->prec) {
				/* prec unset, sibling is unvisited */
				branch = 0;
				break;
			}
		}

		/* we are the last visited children, go on to parent. */
		if(branch) build_precedence_branch(stt, o, tail);
	}
}

/* calculate class precedence */
void stt_build_precedence(stt_ctx_t *stt)
{
	long i;
	stt_var_t *var;
	stt_obj_t *tail;

	stt_obj_t *o, *iter;

	stt->class_precedence = NULL;
	tail = NULL;

	/* set prec pointer of all classes to NULL. (unvisited) */
	for(i=0; i<stt->class_hash->ns; i++) {
		for(var = stt->class_hash->hash[i]; var; var = var->next) {
			if(var->value) {
				var->value->d.clos_class->prec = NULL;
			}
		}
	}

	/* iterate over stt->class_hash
	 * the classes that don't have children. */
	for(i=0; i<stt->class_hash->ns; i++) {
		for(var = stt->class_hash->hash[i]; var; var = var->next) {
			if(var->value) {
				if(var->value->d.clos_class->children->type == STT_ST_NIL) {
					build_precedence_branch(stt, var->value, &tail);
				}
			}
		}
	}

	/* set precedence scores */
	iter = stt->class_precedence; i = 0;
	while(stt_nextparm(&o, &iter)) {
		o->d.clos_class->prec_score = i++;
	}
}

/* calculate and update data for method */
stt_obj_t *stt_method_calculate(stt_varctx_t *ctx, stt_obj_t *name, stt_obj_t *method)
{
	stt_obj_t *generic;
	stt_obj_t **mlist;
	int order = 1;
	stt_var_t *var;

	assert(name->type == STT_ST_SYMBOL);

	/* check whether a generic already exists by this name */
	var = stt_varctx_get_s(ctx, name->d.symbol.sdbm, name->d.symbol.id);
	if(var) {
		int i;

		/* there's a binding for this name */
		if(var->value->type != STT_ST_GENERIC)
			return stt_create_error(ctx->stt, STT_SE_TYPE, "method_calculate: name is bound but not a generic");
		generic = var->value;

		/* clear dispatch table */
		generic->d.clos_generic.dispatch_fill = 0;

		for(i = 0; i<generic->d.clos_generic.dispatch_size; i++) {
			generic->d.clos_generic.dispatch[i].method = NULL;
			free(generic->d.clos_generic.dispatch[i].signature);
			generic->d.clos_generic.dispatch[i].signature = NULL;
		}

	} else {
		/* name unbound, create generic */
		generic = stt_create_generic(ctx->stt);

		/* bind name in global ctx */
		while(ctx->parent) ctx = ctx->parent;
		stt_varctx_set_s(ctx, name->d.symbol.sdbm, name->d.symbol.id, generic);
	}


	/* determine which method list should this be inserted to */
	switch(method->d.clos_method.place) {
	case CL_METHOD_BEFORE:
		mlist = &generic->d.clos_generic.m_before;
		break;
	case CL_METHOD_AFTER:
		mlist = &generic->d.clos_generic.m_after;
		order = -1; /* :after list is sorted the other way */
		break;
	case CL_METHOD_AROUND:
		mlist = &generic->d.clos_generic.m_around;
		break;
	default:
	case CL_METHOD_PRIMARY:
		mlist = &generic->d.clos_generic.m_primary;
		break;
	}

	/* is target list empty? */
	if((*mlist)->type == STT_ST_NIL) {
		/* if so, put this as first element and exit. */

		*mlist = stt_create_cons(ctx->stt, method, stt_create_nil(ctx->stt));
		stt_update_generic_nexts(ctx->stt, generic);
		return stt_create_t(ctx->stt);
	}

	if(order * stt_compare_specificity(ctx->stt, method, (*mlist)->d.cons.car) > 0) {
		/* more specific than first element in list */

		*mlist = stt_create_cons(ctx->stt, method, *mlist);
		stt_update_generic_nexts(ctx->stt, generic);
		return stt_create_t(ctx->stt);
	}

	/* skip *mlist until element before list end
	 * or until next element is less specific */
	while((*mlist)->d.cons.cdr->type == STT_ST_CONS && order * stt_compare_specificity(ctx->stt, method, (*mlist)->d.cons.cdr->d.cons.car) < 0) {
		mlist = &(*mlist)->d.cons.cdr;
	}

	if((*mlist)->d.cons.cdr->type == STT_ST_NIL) {
		/* append after list. */

		(*mlist)->d.cons.cdr = stt_create_cons(ctx->stt, method, stt_create_nil(ctx->stt));
		stt_update_generic_nexts(ctx->stt, generic);
		return stt_create_t(ctx->stt);
	}

	if(stt_compare_specificity(ctx->stt, method, (*mlist)->d.cons.cdr->d.cons.car) == 0) {
		/* equally specific to next element
		 * replace it. */

		(*mlist)->d.cons.cdr->d.cons.car = method;
		stt_update_generic_nexts(ctx->stt, generic);
		return stt_create_t(ctx->stt);
	}

	/* next element is less specific, insert before. */

	(*mlist)->d.cons.cdr = stt_create_cons(ctx->stt, method, (*mlist)->d.cons.cdr);
	stt_update_generic_nexts(ctx->stt, generic);
	return stt_create_t(ctx->stt);
}

/* tries to find a valid next method in lst */
static stt_obj_t *find_next_method(stt_ctx_t *stt, stt_obj_t *method, stt_obj_t *lst)
{
	stt_obj_t *item, *iter;

	assert(method->type == STT_ST_METHOD);

	iter = lst;
	while(stt_nextparm(&item, &iter)) {
		if(stt_is_next_method(stt, method, item))
			return item;
	}

	return NULL;
}

/* update next pointers for all methods in a generic */
void stt_update_generic_nexts(stt_ctx_t *stt, stt_obj_t *generic)
{
	stt_obj_t *item, *iter;

	assert(generic->type == STT_ST_GENERIC);

	/* primary methods */
	iter = generic->d.clos_generic.m_primary;
	while(stt_nextparm(&item, &iter)) {
		if(iter->type == STT_ST_CONS)
			item->d.clos_method.next = find_next_method(stt, item, iter);
		else
			item->d.clos_method.next = NULL;
	}

	/* before methods are now unoptimized */
	iter = generic->d.clos_generic.m_before;
	while(stt_nextparm(&item, &iter)) {
		if(iter->type == STT_ST_CONS)
			item->d.clos_method.next = iter->d.cons.car;
		else
			item->d.clos_method.next = NULL;
	}

	/* after methods are now unoptimized */
	iter = generic->d.clos_generic.m_after;
	while(stt_nextparm(&item, &iter)) {
		if(iter->type == STT_ST_CONS)
			item->d.clos_method.next = iter->d.cons.car;
		else
			item->d.clos_method.next = NULL;
	}

	/* around methods */
	iter = generic->d.clos_generic.m_around;
	while(stt_nextparm(&item, &iter)) {
		if(iter->type == STT_ST_CONS) {
			item->d.clos_method.next = find_next_method(stt, item, iter);
			if(item->d.clos_method.next == NULL)
				item->d.clos_method.next = find_next_method(stt, item, generic->d.clos_generic.m_primary);
		} else {
			/* last around method is followed by first primary method */
			if(generic->d.clos_generic.m_primary->type == STT_ST_CONS) {
				item->d.clos_method.next = generic->d.clos_generic.m_primary->d.cons.car;
			} else {
				item->d.clos_method.next = NULL;
			}
		}
	}
}

/* compare specificity of two methods
 * returns
 *  <0   if a is less specific than b
 *  0	if a has the same signature as b
 *  >0   if a is more specific than b
 */
int stt_compare_specificity(stt_ctx_t *stt, const stt_obj_t *a, const stt_obj_t *b)
{
	stt_method_arg_t *aa, *ab;
	int arga = -1, argb = -1;

	assert(a->type == STT_ST_METHOD);
	assert(b->type == STT_ST_METHOD);

	for(;;) {
		arga ++;
		argb ++;

		if(arga == a->d.clos_method.argn) {
			/* "A" out of arguments. */

			/* if "B" is too, they're considered equal */
			if(argb == b->d.clos_method.argn)
				return 0;

			/* otherwise, "A" is less specific. */
			return -1;
		}

		/* if "B" is out of arguments, "A" is more specific. */
		if(argb == b->d.clos_method.argn)
			return 1;

		aa = a->d.clos_method.args + arga;
		ab = b->d.clos_method.args + argb;

		if(aa->value) {
			/* "A" has value filter */

			/* if "B" hasn't, "A" is more specific */
			if(!ab->value)
				return 1;

			/* if "A" value == "B" value, continue to next args */
			if(stt_isoform(aa->value, ab->value))
				continue;

			/* they both have filters, for different values
			 * order them arbitrarily */
			return -1;
		}

		if(ab->value) {
			/* "B" has value filter, where "A" hasn't
			 * "B" is more specific. */
			return -1;
		}

		if(aa->type > -1) {
			/* "A" has a type specified */

			/* if "B" has the same type specified, continue. */
			if(aa->type == ab->type)
				continue;

			/* if "B" has no type, "A" is more specific. */
			if(ab->type == -1)
				return 1;

			if(aa->type > ST_TYPE_LAST_MARK) {
				/* "A" has a class type */

				/* if "B" has an atomic type, "A" is more specific. */
				if(ab->type < ST_TYPE_LAST_MARK)
					return 1;

				/* both have class types, compare class specificity. */
				if(stt->typedb[aa->type].cl->d.clos_class->prec_score < stt->typedb[ab->type].cl->d.clos_class->prec_score) {
					/* if "A" is before "B" in the class precedence list,
					 * "A" is more specific. */
					return 1;
				} else {
					/* if "A" is after "B", "A" is less specific. */
					return -1;
				}

			} else {
				/* "A" has an atomic type */

				/* "B" is considered more specific if it has a class type
				 * (> ST_TYPE_LAST_MARK) or has a larger type ID. */

				return ab->type > aa->type ? -1 : 1;
			}
		}

		if(ab->type > -1) {
			/* "B" has a type specified, but "A" hasn't.
			 * "B" is more specific */

			return -1;
		}

		/* both args unspecified, continue. */
	}
}

/* returns true if A class is a generalisation of, or equal to, B class */
int stt_is_generalisation(const stt_obj_t *a, const stt_obj_t *b)
{
	stt_obj_t *item, *iter;

	assert(a->type == STT_ST_CLASS);
	assert(b->type == STT_ST_CLASS);

	if(a == b) return 1;

	iter = b->d.clos_class->ancestry;
	while(stt_nextparm(&item, &iter)) {
		if(item == a) return 1;
	}

	return 0;
}

/* returns true if B method is a possible next-method of A method. */
int stt_is_next_method(stt_ctx_t *stt, const stt_obj_t *a, const stt_obj_t *b)
{
	stt_method_arg_t *aa, *ab;
	int arga = -1, argb = -1;

	assert(a->type == STT_ST_METHOD);
	assert(b->type == STT_ST_METHOD);

	for(;;) {
		arga ++;
		argb ++;

		if(arga == a->d.clos_method.argn) {
			/* "A" out of arguments. */

			/* if "B" is too, then B is a next-method */
			if(argb == b->d.clos_method.argn)
				return 1;

			/* otherwise, it's not. */
			return 0;
		}

		if(argb == b->d.clos_method.argn) {
			/* "B" is out of arguments.
			 * "A" has more, so "B" is not a next-method for "A". */
			return 0;
		}

		aa = a->d.clos_method.args + arga;
		ab = b->d.clos_method.args + argb;

		if(aa->value) {
			/* "A" has a value filter */

			if(ab->value) {
				/* "B" has as well. */

				if(stt_isoform(aa->value, ab->value)) {
					/* the two values are equal, continue */
					continue;
				}

				/* the two values aren't equal. fail */
				return 0;
			}

			/* "B" has no value filter. */

			/* if "B" has no type specialization, then it's ok */
			if(ab->type == -1)
				continue;

			/* it has type specialization. it has to be the same type
			 * or a generalisation */
			if(aa->value->type == ab->type)
				continue;

			/* fail if "A" value filter isn't an instance */
			if(aa->value->type != STT_ST_OBJECT) return 0;

			/* fail if "B" type is atomic */
			if(ab->type < ST_TYPE_LAST_MARK) return 0;

			if(stt_is_generalisation(stt->typedb[ab->type].cl, aa->value->d.clos_object.type))
				continue;

			return 0;
		}

		if(aa->type != -1) {
			/* "A" has a type specialization */

			if(ab->value) {
				/* if "B" has a value filter, it has to be the same type
				 * or a generalisation. */

				if(ab->value->type == aa->type)
					continue;

				if(ab->value->type != STT_ST_OBJECT) return 0;
				if(aa->type < ST_TYPE_LAST_MARK) return 0;
				if(stt_is_generalisation(ab->value->d.clos_object.type, stt->typedb[aa->type].cl))
					continue;

				return 0;
			}

			/* if "B" hasn't a type specialization, continue */
			if(ab->type == -1)
				continue;

			/* otherwise, it has to be the same type or a generalisation */
			if(aa->type == ab->type)
				continue;

			/* fail if either type is atomic */
			if(aa->type < ST_TYPE_LAST_MARK || ab->type < ST_TYPE_LAST_MARK) return 0;

			if(stt_is_generalisation(stt->typedb[ab->type].cl, stt->typedb[aa->type].cl))
				continue;

			return 0;
		}

		/* both arguments are unspecialized, continue */
	}
}

/* checks if a method is applicable to an argument array.
 * (last element NULL)
 * returns first argument in the array unaccounted for
 * (this doesn't care about misc arguments), or NULL.
 */
stt_obj_t **stt_is_applicable(stt_ctx_t *stt, const stt_obj_t *method, stt_obj_t **args)
{
	int arg = 0;
	stt_obj_t **iter;

	assert(method->type == STT_ST_METHOD);

	if(method->d.clos_method.argn == 0)
		return args;

	for(iter = args; *iter; iter++) {
		stt_method_arg_t *ma = method->d.clos_method.args + arg;

		/* eql filter */
		if(ma->value) {
			if(stt_isoform(ma->value, *iter)) {
				arg++;
				if(arg == method->d.clos_method.argn) return iter;
				continue;
			} else {
				return NULL;
			}
		}

		/* type specialization */
		if(ma->type != -1) {
			if(ma->type < ST_TYPE_LAST_MARK) {
				/* atomic type */
				if((*iter)->type == ma->type) {
					arg++;
					if(arg == method->d.clos_method.argn) return iter;
					continue;
				} else {
					return NULL;
				}
			}

			/* class type, arg has to be an instance */
			if((*iter)->type != STT_ST_OBJECT)
				return NULL;

			if(!stt_is_generalisation(stt->typedb[ma->type].cl, (*iter)->d.clos_object.type))
				return NULL;
		}

		arg++;
		if(arg == method->d.clos_method.argn) {
			/* at this place, we don't care of misc arguments. */
			return iter;
		}
	}

	/* more specific arguments than elements in the argument list. */
	return NULL;
}

/* expand/zero dispatch table */
void stt_expand_dispatch(stt_obj_t *generic)
{
	int i;
	for(i=0; i<generic->d.clos_generic.dispatch_size; i++)
		free(generic->d.clos_generic.dispatch[i].signature);

	generic->d.clos_generic.dispatch_fill = 0;
	generic->d.clos_generic.dispatch_size *= 2;
	generic->d.clos_generic.dispatch_limit = (generic->d.clos_generic.dispatch_size * DISPATCH_THRESHOLD) >> 8;
	free(generic->d.clos_generic.dispatch);
	generic->d.clos_generic.dispatch = (stt_dispatch_entry_t *) stt_calloc(generic->d.clos_generic.dispatch_size, sizeof(stt_dispatch_entry_t));
}

/* do method dispatch
 * argeval should be an array of stt_obj_t * (last one NULL) of pre-evaluated arguments.
 * returns the STT_ST_METHOD the call should use, or NULL
 * if ignore_around is 1, around methods are ignored
 */
stt_obj_t *stt_method_dispatch(stt_ctx_t *stt, stt_obj_t *generic, stt_obj_t **argeval, const int ignore_around)
{
	int hashval = 0;
	int fail, size, row;
	stt_dispatch_entry_t *table;
	stt_obj_t **iter;

	assert(generic->type == STT_ST_GENERIC);

	/* compute hash value */
	for(iter = argeval; *iter; iter++)
		hashval += stt->typedb[(*iter)->type].typeseed;

	table = generic->d.clos_generic.dispatch;
	size = generic->d.clos_generic.dispatch_size;
	row = hashval & (size-1);

	/* find row in dispatch table */
	for(;;) {
		if(table[row].method == NULL) break;	/* no hit */
		if(table[row].hash == hashval && table[row].ignore_around == ignore_around) {
			/* hash values match, check signature */

			stt_dispatch_signature_t *ds = table[row].signature - 1;
			stt_obj_t **arg = argeval - 1;
			fail = 0;
			for(;;) {
				ds++;
				arg++;

				/* if out of signature, it's a match */
				if(ds->type == -1) {
					break;
				}

				/* if out of arguments, no match. */
				if(*arg == NULL) {
					fail = 1;
					break;
				}

				/* if specified, check for value match */
/*				if(ds->value) {
						if(!stt_isoform(ds->value, *arg)) {
							fail = 1;
							break;
						}
*/
					/* no need to check type in this case */
/*					continue;
				}*/

				if(ds->type < ST_TYPE_LAST_MARK) {
					/* atomic type */
					if(ds->type != (*arg)->type) {
						fail = 1;
						break;
					}
				} else {
					/* class type, arg has to be an instance */
					if((*arg)->type != STT_ST_OBJECT) {
						fail = 1;
						break;
					}

					if((*arg)->d.clos_object.type->d.clos_class->typeid != ds->type) {
						fail = 1;
						break;
					}
				}
			}

			if(!fail) {
				/* found a match from dispatch table */
				return table[row].method;
			}
		}

		row = (row + 1) & (size - 1);
	}

	/* no result, calculate method by hand */
	{
		stt_obj_t *item, *mlist;
		if(!ignore_around) {
			/* look for most specific applicable around method. */
			mlist = generic->d.clos_generic.m_around;
			while(stt_nextparm(&item, &mlist)) {
				stt_obj_t **misc = stt_is_applicable(stt, item, argeval);

				if(misc) {
					if(generic->d.clos_generic.dispatch_fill == generic->d.clos_generic.dispatch_limit) {
						stt_expand_dispatch(generic);
						stt_write_dispatch(generic, hashval & (size*2 - 1), hashval, argeval, item, ignore_around);
					} else {
						stt_write_dispatch(generic, row, hashval, argeval, item, ignore_around);
					}

					return item;
				}
			}
		}


		/* no around method found, look in primary methods */
		mlist = generic->d.clos_generic.m_primary;
		while(stt_nextparm(&item, &mlist)) {
			stt_obj_t **misc = stt_is_applicable(stt, item, argeval);

			if(misc) {
				if(generic->d.clos_generic.dispatch_fill == generic->d.clos_generic.dispatch_limit) {
					stt_expand_dispatch(generic);
					stt_write_dispatch(generic, hashval & (size*2 - 1), hashval, argeval, item, ignore_around);
				} else {
					stt_write_dispatch(generic, row, hashval, argeval, item, ignore_around);
				}

				return item;
			}
		}
	}
	/* nothing found. */
	return NULL;
}

/* write row into dispatch table
 * finds first empty row starting from <row>
 * returns that row, after filling it in.
 */
int stt_write_dispatch(stt_obj_t *generic, int row, int hashval, stt_obj_t **argeval, stt_obj_t *method, const int ignore_around)
{
	int i, size, arglen;
	stt_dispatch_entry_t *table;

	assert(generic->type == STT_ST_GENERIC);

	table = generic->d.clos_generic.dispatch;
	size = generic->d.clos_generic.dispatch_size;
	for(arglen = 0; argeval[arglen]; arglen++);

	while(table[row].method) {
		row = (row + 1) & (size - 1);
	}

	table[row].hash = hashval;
	table[row].method = method;

	table[row].signature = (stt_dispatch_signature_t *) stt_calloc(arglen + 1, sizeof(stt_dispatch_signature_t));
	table[row].ignore_around = ignore_around;


	for(i=0; i<arglen; i++) {
		if(argeval[i]->type == STT_ST_OBJECT)
			table[row].signature[i].type = argeval[i]->d.clos_object.type->d.clos_class->typeid;
		else
			table[row].signature[i].type = argeval[i]->type;
	}

	/* append a -1 type after last element. */
	table[row].signature[i].type = -1;

	generic->d.clos_generic.dispatch_fill ++;

	return row;
}


/*************************************** DICTIONARY ***/


/* rehash dictionary */
void stt_dictrehash(stt_ctx_t *stt, stt_obj_t *dict)
{
	unsigned long i;
	unsigned long ns = dict->d.dict.size * 2;
	unsigned long nm = ns - 1;
	stt_dictentry_t *ndata = (stt_dictentry_t *) stt_calloc(ns, sizeof(stt_dictentry_t));
	stt_dictentry_t *odata = dict->d.dict.data;

	dict->d.dict.fill = 0;
	for(i=0; i<dict->d.dict.size; i++) {
		if(odata[i].name != NULL && odata[i].name != (stt_obj_t *) 1)
			dict->d.dict.fill += stt_dictput_(ndata, nm, odata[i].hash, odata[i].name, odata[i].value);
	}
	free(odata);
	dict->d.dict.data = ndata;
	dict->d.dict.size = ns;
	dict->d.dict.mask = nm;
	dict->d.dict.threshold = (ns * stt->dict_threshold) >> 8;
}

/* put value into dictionary (helper function) */
int stt_dictput_(stt_dictentry_t *data, unsigned int mask, unsigned long keyhash, stt_obj_t *key, stt_obj_t *value)
{
	unsigned int slot = keyhash & mask;
	unsigned int s = slot;
	for(;;) {
		if(!data[s].name || ((stt_intptr_t) data[s].name) == 1) {
			data[s].name = key;
			data[s].value = value;
			data[s].hash = keyhash;
			return 1;
		}
		if(data[s].hash == keyhash && stt_isoform(data[s].name, key)) {
			data[s].value = value;
			return 0;
		}

		s = (s + 1) & mask;
		assert(s != slot);
	}
}

/* put value into dictionary */
void stt_dictput(stt_ctx_t *stt, stt_obj_t *dict, unsigned long keyhash, stt_obj_t *key, stt_obj_t *value)
{
	if(dict->d.dict.fill == dict->d.dict.threshold)
		stt_dictrehash(stt, dict);
	dict->d.dict.fill += stt_dictput_(dict->d.dict.data, dict->d.dict.mask, keyhash, key, value);
}

/* get value from dictionary */
stt_obj_t *stt_dictget(const stt_obj_t *dict, const stt_obj_t *key)
{
	unsigned long int keyhash = stt_hashfunc(key);
	unsigned long slot = keyhash & dict->d.dict.mask;
	unsigned long s = slot;
	stt_dictentry_t *data = dict->d.dict.data;
	unsigned long mask = dict->d.dict.mask;
	for(;;) {
		if(!data[s].name) return NULL;
		if(((stt_intptr_t) data[s].name) != 1 && data[s].hash == keyhash && stt_isoform(data[s].name, key))
			return data[s].value;

		s = (s + 1) & mask;
		if(s == slot) return NULL;
	}
}

/* delete item from dictionary */
int stt_dictdel(stt_obj_t *dict, const stt_obj_t *key)
{
	unsigned long int keyhash = stt_hashfunc(key);
	unsigned long slot = keyhash & dict->d.dict.mask;
	unsigned long s = slot;
	stt_dictentry_t *data = dict->d.dict.data;
	unsigned long mask = dict->d.dict.mask;
	for(;;) {
		if(!data[s].name) return 0;
		if((stt_intptr_t) data[s].name != 1 && data[s].hash == keyhash && stt_isoform(data[s].name, key)) {
			data[s].name = (stt_obj_t *) 1;
			return 1;
		}

		s = (s + 1) & mask;
		if(s == slot) return 0;
	}
}

/****** utility ******/

/* creates a new string builder */
stt_strbuild_t stt_strbuild_new()
{
	stt_strbuild_t empty = { NULL, 0, 0 };
	return empty;
}

/* append a character */
void stt_strbuild_addch(stt_strbuild_t *token, int ch)
{
	if(!token->str) {
		token->str = (char *) stt_malloc(8);
		token->len = 2;
		token->buf = 8;
		token->str[0] = ch;
		token->str[1] = 0;
		return;
	}

	if(token->len == token->buf) {
		token->str = (char *) stt_realloc(token->str, token->buf * 2);
		token->buf *= 2;
	}

	token->str[token->len-1] = ch;
	token->str[token->len++] = 0;
}

/* append len characters from a string */
void stt_strbuild_addstr(stt_strbuild_t *token, const char *str, size_t len)
{
	size_t buf;

	if(!token->str) {
		buf = len + 1;
		if(buf < 8) buf = 8;
		token->str = (char *) stt_malloc(buf);
		token->len = len + 1;
		token->buf = buf;
		memcpy(token->str, str, len);
		token->str[len] = 0;
		return;
	}

	if(token->len + len >= token->buf) {
		buf = token->buf;
		do { buf *= 2; } while(token->len + len >= buf);
		token->str = (char *) stt_realloc(token->str, buf);
		token->buf = buf;
	}

	memcpy(token->str + token->len - 1, str, len);
	token->len += len;
	token->str[token->len - 1] = 0;
}

/* define a new cookie type (returns id) */
int stt_new_cookie_type()
{
	return cookie_type_cnt ++;
}

/* quote an object for passing.. returns (quote obj) */
stt_obj_t *stt_quoteobj(stt_ctx_t *stt, stt_obj_t *obj)
{
	return stt_create_cons(stt, stt->quote_builtin, stt_create_cons(stt, obj, stt_create_nil(stt)));
}
