#ifndef _ST_TYPES_H
#define _ST_TYPES_H

#include <stdio.h>
#include <estutter/stt_config.h>
#include <stdarg.h>
#include <stddef.h>

typedef struct stt_ctx_s stt_ctx_t;

/* this'd better be 64 bits but there's no long long in c89 */
typedef long stt_serial_t;

/* This most often works; at least works if sizeof(ptrdiff_t) >= sizeof(void *)
   (we check that on init) */
typedef ptrdiff_t stt_intptr_t;


/* primitive object types */
typedef enum {
	STT_ST_NIL, /* array depends on this being zero */
	STT_ST_INT,
	STT_ST_NUM,
	STT_ST_CHAR,
	STT_ST_T,
	STT_ST_STRING,
	STT_ST_BLOB,

	STT_ST_COOKIE,
	STT_ST_EOF,

	STT_TYPE_ATOM_MARK, /* no types before this have references to other objects */

	STT_ST_BUILTIN,
	STT_ST_FUNC,
	STT_ST_ERROR,
	STT_ST_DEFER,
	STT_ST_ARRAY,
	STT_ST_VARCTX,
	STT_ST_DICT,
	STT_ST_STREAM,

	STT_ST_CLASS,
	STT_ST_SLOT,
	STT_ST_OBJECT,
	STT_ST_GENERIC,
	STT_ST_METHOD,
	STT_ST_DISPATCH,

	STT_TYPE_NOEVAL_MARK, /* all types before this evaluate to themselves */

	STT_ST_CONS,
	STT_ST_SYMBOL,
	STT_ST_VARREF,
	STT_ST_ANNOT,

	ST_TYPE_LAST_MARK /* first non-used type ID */
} stt_type_t;

/* forward declaration of generic object type */
typedef struct stt_obj_s stt_obj_t;

typedef struct {
	int typeseed;	/* seed for a type */
	stt_obj_t *cl;		/* class object, if applicable */
} stt_typeinfo_t;

/* CONS object */
typedef struct {
	stt_obj_t *car;
	stt_obj_t *cdr;
} stt_cons_t;

/* ANNOT object */
typedef struct {
	stt_obj_t *val;
	stt_obj_t *annot;
} stt_annot_t;

/* Number object */
typedef struct {
	double value;
} stt_num_t;

/* Integer object */
typedef struct {
	long value;
} stt_int_t;

typedef struct stt_var_s stt_var_t;

/* Symbol object */
typedef struct {
	char *name;
	unsigned long sdbm;
	unsigned long id;
	int colon;
} stt_symbol_t;

/* Stack variable reference */
typedef struct {
	ptrdiff_t offset; /* added to BP. in bytes (so cell*sizeof(void*)) */
} stt_varref_t;

/* String object */
typedef struct {
	char *value;
} stt_string_t;

/* Blob object */
typedef struct {
	unsigned char *data;
	size_t size;
} stt_blob_t;

/* Array object */
typedef struct {
	int dim; /* number of dimensions */
	long size[STT_ARRAY_MAXDIMS]; /* size of each dimension */
	stt_obj_t **data; /* data */
} stt_array_t;

/* Character object */
typedef struct {
	unsigned char ch;
} stt_char_t;

/* Entry in a dictionary */
typedef struct {
	unsigned long hash;
	stt_obj_t *name;
	stt_obj_t *value;
} stt_dictentry_t;

/* Dictionary object */
typedef struct {
	unsigned long size; /* dictionary size (power of 2) */
	unsigned long mask; /* size - 1 */
	unsigned long fill; /* number of slots occupied */
	unsigned long threshold; /* occupation threshold (surpassing which a rehash occurs) */
	stt_dictentry_t *data;
} stt_dict_t;

/* Stream types */
enum {
	STT_STREAM_FP,
	STT_STREAM_FUNC,
	STT_STREAM_C
};

/* Stream object */
typedef struct {
	int type;
	int unread; /* a single character put back in the stream */

	/* parsing position */
	long line;

	union {
		FILE *fp;    /* STT_STREAM_FP */
		stt_obj_t *func; /* STT_STREAM_FUNC */
		struct {     /* STT_STREAM_C */
			void *data;
			int type;
			int (*reader)(void **data, const int type); /* read and return 1 character */
			int (*writer)(int c, void **data, const int type); /* write 1 character */
			size_t (*readmore)(void **data, const int type, void *buf, const size_t len);
			size_t (*writemore)(void **data, const int type, void *buf, const size_t len);
			size_t (*vprintf)(void **data, const int type, const char *fmt, va_list ap);
			void (*destroy)(stt_obj_t *o);
		} c;
	} s;
} stt_stream_t;

struct stt_varctx_s; /* defined in memory.h */
typedef struct stt_varctx_s stt_varctx_t;

/* prototype for builtin handler */
typedef stt_obj_t* (* stt_builtin_handler) (stt_varctx_t *, stt_obj_t *, void *, const int tail);

/* Builtin function object */
typedef struct {
	stt_builtin_handler handler; /* function handler */
	void *data; /* user data passed to handler */
	char *symbol; /* first name this builtin got bound to.. */
	unsigned long symbolid; /* first symbol ID this builtin got bound to */
	char printprefix; /* when printing, use {prefix}(list) instead of (builtin list) */
	int nosubst; /* substitution defuse flag */
} stt_builtin_t;

/* Function object */
typedef struct {
	stt_obj_t *parm; /* list of function parameters */
	stt_obj_t *exp; /* function body: list of expressions */
	stt_obj_t *docstr; /* docstring */
	char *symbol; /* first name this func got bound to.. */
	int macro;	/* 1 for macro, 2 for r-macro
				 * 4 for symbolic-macro, 5 for r-symbolic-macro */
	stt_obj_t *closure; /* if this is a closure, reference to a STT_ST_VARCTX object */
	int trace; /* boolean */
} stt_func_t;

/* Function call deferral object */
typedef struct {
	stt_obj_t *func; /* user function object */
	stt_obj_t *args; /* list of arguments */
	stt_obj_t *obj; /* s-expression of function call */
} stt_defer_t;

/* Error object */
typedef struct {
	int code; /* error code */
	stt_obj_t *obj; /* thrown object */
	char *msg; /* error message */
	char *file; /* location */
	stt_obj_t *trace; /* error trace */
	long line;
} stt_error_t;

/* Cookie object */
typedef struct {
	int type; /* cookie type */
	void *ptr;
	void (* destroy) (stt_obj_t *o);
} stt_cookie_t;

/* Encapsulated stt_varctx_t */
typedef struct {
	stt_varctx_t *ctx;
} stt_varctx_enc_t;

/****************************** CLOS TYPES ***/

/* slot assignment */
typedef struct {
	stt_obj_t *stor;	/* STT_ST_CLASS on static, or NULL */
	stt_obj_t *slot;
	size_t num;
} stt_slotasg_t;

/* CLOS class */
typedef struct {
	stt_obj_t *name;	/* symbol */
	int typeid;	/* type ID */

	stt_obj_t *parents;	/* list of parents (STT_ST_CLASS) */
	stt_obj_t *slots;	/* list of own slots (STT_ST_SLOT) */

	stt_obj_t *children;	/* list of children (STT_ST_CLASS) */

	stt_obj_t *prec;	/* cons in precedence list */
	int prec_score;	/* position in precedence list */

	stt_obj_t *ancestry;	/* ancestors sorted by precedence */

	stt_slotasg_t *allslots;	/* list of all slots */
	size_t allslotsnum;	/* number of all slots */

	stt_varctx_t *initarg;	/* map initarg->slot_asg (idx) */
	stt_varctx_t *slotname;	/* map slotname->slot_asg (idx) */

	stt_obj_t **data;	/* storage of static data */
	size_t data_size;

	size_t instdata;	/* data size for instances */
} stt_class_t;

/* CLOS object */
typedef struct {
	stt_obj_t *type;	/* STT_ST_CLASS */

	stt_obj_t **data;	/* data storage */
	size_t data_size;
} stt_object_t;

/* slot allication type */
typedef enum {
	CL_ALLOC_CLASS,
	CL_ALLOC_INSTANCE
} stt_slot_allocation_t;

/* CLOS slot */
typedef struct {
	stt_obj_t * name;	/* name: symbol */
	stt_obj_t * accessor;	/* name of accessor, or NULL */
	stt_obj_t * writer;	/* name of writer, or NULL */
	stt_obj_t * reader;	/* name of reader, or NULL */
	stt_obj_t * initarg;	/* name of initarg, or NULL */

	stt_slot_allocation_t alloc;	/* type of slot allocation */
	stt_obj_t *initform;	/* initializer expression */
} stt_slot_t;

/* signature entry in dispatch table */
typedef struct {
	int type;		/* type id used in call, -1 after last */
	stt_obj_t *value;	/* instance used in call, or NULL */
} stt_dispatch_signature_t;

/* entry in method dispatch table */
typedef struct {
	int hash;	
	stt_obj_t *method;	/* method mapped to call, or NULL */
	stt_dispatch_signature_t *signature;	/* signature of call */
	int ignore_around;
} stt_dispatch_entry_t;

/* CLOS generic function */
typedef struct {
	stt_obj_t * m_primary;  /* list of primary methods, most specific to least */
	stt_obj_t * m_before;   /* :before methods, most specific to least */
	stt_obj_t * m_after;	/* :after methods, least specific to most */
	stt_obj_t * m_around;   /* :around methods, most specific to least */

	stt_dispatch_entry_t *dispatch;	/* method dispatch table */
	int dispatch_size;			/* entries in dispatch table */
	int dispatch_fill;			/* filled entries */
	int dispatch_limit;			/* limit of fill before expansion */
} stt_generic_t;

/* possible places for a method */
typedef enum {
	CL_METHOD_BEFORE,
	CL_METHOD_PRIMARY,
	CL_METHOD_AFTER,
	CL_METHOD_AROUND
} stt_method_place_t;

/* method argument definition */
typedef struct {
	stt_obj_t *name;	/* argument name (symbol) */
	int type;		/* type ID, or -1 */
	stt_obj_t *value;	/* value filter (eql ...) or NULL */
} stt_method_arg_t;

/* CLOS method */
typedef struct {
	stt_method_place_t place;		/* method type (before, after, etc) */
	stt_method_arg_t *args;		/* array of specific arguments */
	int argn;				/* number of specific arguments */

	stt_obj_t *miscargs;		/* list of misc arguments */
	stt_obj_t *code;			/* method code */
	stt_obj_t *next;			/* next specific method, or NULL */
} stt_method_t;

/* Method dispatch state */
typedef struct {
	stt_obj_t **argeval;		/* array of evaluated arguments (FUNC_MAXPARMS + 1 long) */
	stt_obj_t *generic;			/* generic */
	stt_obj_t *method;			/* selected method */
} stt_dispatch_t;

/*********************** INTERNAL TYPES ***/

/* type-dependent data is contained in this union */
union stt_obj_data_u {
	stt_cons_t cons;
	stt_num_t num;
	stt_symbol_t symbol;
	stt_varref_t varref;
	stt_error_t error;
	stt_builtin_t builtin;
	stt_func_t func;
	stt_string_t string;
	stt_int_t integer;
	stt_defer_t defer;
	stt_array_t array;
	stt_varctx_enc_t varctx;
	stt_dict_t dict;
	stt_stream_t stream;
	stt_cookie_t cookie;
	stt_char_t ch;
	stt_annot_t annot;
	stt_blob_t blob;

	stt_class_t *clos_class;
	stt_slot_t clos_slot;
	stt_object_t clos_object;
	stt_generic_t clos_generic;
	stt_method_t clos_method;
	stt_dispatch_t clos_dispatch;
};

typedef struct weakptr_s stt_weakptr_t;
struct weakptr_s {
	stt_obj_t *host;			 /* link host */
	stt_serial_t serial; /* serial of link host */
	stt_obj_t **ptr;			 /* link location */
	stt_weakptr_t *next;
};

/* generic object of the language */
struct stt_obj_s {
	stt_serial_t serial; /* serial number */

	stt_type_t type; /* object type */
	char *file; /* point of definition */
	long line;
	union stt_obj_data_u d; /* type-dependent data contained in the object */
	stt_weakptr_t *weak;  /* list of weak pointers pointing here */

	unsigned long int mark; /* used by the GC */
};

typedef struct {
	char *str;
	size_t len;
	size_t buf;
} stt_strbuild_t;


/* dispatch table for object follow */
typedef void (*stt_gc_follow_dispatcher_t)(stt_ctx_t *stt, stt_obj_t *obj);

/* Execution context */
struct stt_ctx_s {

	/*** public configuration (may be modified by the caller) ***/
	stt_strbuild_t refpath; /* current refpath */
	long trace_depth;
	long recursion_limit; /* levels of recursion left */
	int macro_annot; /* whether to insert -annot around macro expansions for debugging */
	unsigned long default_dict_size; /* default size for a dictionary */
	unsigned long dict_threshold;    /* dictionary rehash threshold (/256) */
	int gc_alloc_limit; /* perform GC each time after this many objects are allocated */
	int gc_verbose; /* verbose GC */
	int gc_disabled;

	void *user_data; /* reserved for caller's use */

	/*** private internal states - should not be modified directly by the caller ***/

	/* eval: parse and config */
	char *load_path; /* current directory for load */
	char *file; /* current file being parsed */
	char *lastcall_file; /* location of last function call */
	size_t lastcall_line;
	size_t parsed_line; /* line number currently being parsed */

	/* eval: language features */
	unsigned long rest_sdbm; /* sdbm of "_REST" */
	unsigned long rest_id; /* unique symbol id of _REST symbol */
	unsigned long dispatch_info_sdbm; /* sdbm of _DISPATCH-INFO symbol */
	unsigned long dispatch_info_id; /* unique symbol id of _DISPATCH-INFO symbol */
	unsigned long lambda_sdbm; /* sdbm of LAMBDA symbol */
	unsigned long lambda_id; /* unique symbol id of LAMBDA symbol */
	unsigned long recurse_sdbm; /* sdbm of <^ (recursion) symbol */
	unsigned long recurse_id; /* unique symbol id of <^ (recursion) symbol */

	/* flat array holding error names (16 byte max each) */
	char *errornames;
	int errorcount; /* index of first free error code */

	/* memory */
	stt_serial_t next_serial; /* object serial number for next allocation */
	int next_typeid; /* next type ID */
	stt_typeinfo_t *typedb; /* type ID -> type info */
	int typedb_size;
	stt_serial_t varctx_id; /* first free stt_varctx_t id */
	unsigned long symbol_id; /* first free symbol id */
	stt_varctx_t *symbol_name_hash; /* global symbol name -> unique id hash */
	stt_varctx_t *class_hash; /* name -> class hash */
	stt_obj_t *class_precedence; /* precedence list of classes */
	stt_obj_t *quote_builtin;

	/* memory: gc */
	unsigned long int gc_mark;
	unsigned long int gc_allocd; /* number of objects allocated since last GC run */
	unsigned long int gc_live; /* number of live objects left */

	int gc_gen2_count, gc_gen3_count; /* generation sweep counters these decrease until 1, then generation is sweeped and counter is reset. */
	stt_gc_follow_dispatcher_t gc_follow_dispatch[ST_TYPE_LAST_MARK];

	/* memory: reusable object caches */
#	ifndef DISABLE_INTCACHE
	stt_obj_t *reusable_ints[STT_REUSABLE_INT_LIMIT];
#	endif
#	ifndef DISABLE_CHARCACHE
	stt_obj_t *reusable_chars[256];
#	endif


	/* memory: gc caches */
		/* list of allocated objects (generation 1) */
	stt_obj_t **gc_list;
	int gc_list_cnt;
	size_t gc_list_size;
	stt_obj_t **gc_list_tail;

			/* list of allocated objects (generation 2) */
	stt_obj_t **gc_gen2;
	int gc_gen2_cnt;
	size_t gc_gen2_size;
	stt_obj_t **gc_gen2_tail;

		/* list of allocated objects (generation 3) */
	stt_obj_t **gc_gen3;
	int gc_gen3_cnt;
	size_t gc_gen3_size;
	stt_obj_t **gc_gen3_tail;

		/* list of protected objects */
	stt_obj_t **gc_prot;
	int gc_prot_cnt;
	size_t gc_prot_size;
	stt_obj_t **gc_prot_tail;

		/* list of miscellaneous pointers to free at shutdown */
	stt_obj_t **miscfree;
	int miscfree_cnt;
	size_t miscfree_size;
	stt_obj_t **miscfree_tail;

		/* recycle pool */
	stt_obj_t **recycle;
	int recycle_cnt;
	size_t recycle_size;
	stt_obj_t **recycle_tail;

} ;


/* initialize reusable object caches */
void stt_objcache_init(stt_ctx_t *stt);


/* creates and returns a new stt_obj_t, zeroed */
stt_obj_t *stt_alloc_obj(stt_ctx_t *stt);

/* creates a new stt_obj_t without updating the records of the GC */
stt_obj_t *stt_alloc_ungc_obj(stt_ctx_t *stt);

/* create a numeric atom */
stt_obj_t *stt_create_num(stt_ctx_t *stt, const double val);

/* create an integer atom */
stt_obj_t *stt_create_int(stt_ctx_t *stt, const long val);

/* create a character atom */
stt_obj_t *stt_create_char(stt_ctx_t *stt, const unsigned char ch);

/* create an EOF atom */
stt_obj_t *stt_create_eof(stt_ctx_t *stt);

/* create a symbolic atom */
stt_obj_t *stt_create_symbol(stt_ctx_t *stt, char *val);

/* create a guaranteed new symbolic atom */
stt_obj_t *stt_create_new_symbol(stt_ctx_t *stt, char *val);

/* create a keyword atom */
stt_obj_t *stt_create_keyword(stt_ctx_t *stt, char *val);

/* create a string */
stt_obj_t *stt_create_string(stt_ctx_t *stt, const char *val);

/* create a string from n bytes */
stt_obj_t *stt_create_string_counted(stt_ctx_t *stt, const char *val, const size_t len);

/* create a string (doesn't stt_strdup val) */
stt_obj_t *stt_create_string_allocd(stt_ctx_t *stt, char *val);

/* create a blob (use pointer, don't copy data) */
stt_obj_t *stt_create_blob(stt_ctx_t *stt, unsigned char *data, size_t size);

/* create a blob (copy data) */
stt_obj_t *stt_create_blob_copy(stt_ctx_t *stt, unsigned char *data, size_t size);

/* create a CONS */
stt_obj_t *stt_create_cons(stt_ctx_t *stt, stt_obj_t *car, stt_obj_t *cdr);

/* create an annot */
stt_obj_t *stt_create_annot(stt_ctx_t *stt, stt_obj_t *val, stt_obj_t *annot);

/* set last element of a list */
void stt_list_setlast(stt_obj_t *lst, stt_obj_t *obj);

/* create an error object */
stt_obj_t *stt_create_error(stt_ctx_t *stt, const int code, const char *message);
stt_obj_t *stt_create_error_with_obj(stt_ctx_t *stt, const int code, const char *message, stt_obj_t *obj);

/* create a builtin object */
stt_obj_t *stt_create_builtin(stt_ctx_t *stt, stt_builtin_handler handler);

/* create a builtin object, with user data */
stt_obj_t *stt_create_builtin_with_data(stt_ctx_t *stt, stt_builtin_handler handler, void *data);

/* return cached T object */
stt_obj_t *stt_create_t(stt_ctx_t *stt);

/* return cached nil object */
stt_obj_t *stt_create_nil(stt_ctx_t *stt);

/* create a T object */
stt_obj_t *stt_create_new_t(stt_ctx_t *stt);

/* create a STT_NIL object */
stt_obj_t *stt_create_new_nil(stt_ctx_t *stt);

/* create a function */
stt_obj_t *stt_create_func(stt_ctx_t *stt, stt_obj_t *prm, stt_obj_t *exp);

/* create a call deferral */
stt_obj_t *stt_create_defer(stt_ctx_t *stt, stt_obj_t *func, stt_obj_t *args, stt_obj_t *obj);

/* create an array */
stt_obj_t *stt_create_array(stt_ctx_t *stt, const int dim, const long *size);

/* create a dictionary */
stt_obj_t *stt_create_dict(stt_ctx_t *stt);

/* create a stream */
stt_obj_t *stt_create_stream(stt_ctx_t *stt, const int type);

/* create encapsulated stt_varctx_t */
stt_obj_t *stt_create_obj_varctx(stt_ctx_t *stt, stt_varctx_t *ctx);

/* create a cookie */
stt_obj_t *stt_create_cookie(stt_ctx_t *stt, const int type, void *ptr);

/* create a class */
stt_obj_t *stt_create_class(stt_ctx_t *stt);

/* create a slot */
stt_obj_t *stt_create_slot(stt_ctx_t *stt);

/* create a object */
stt_obj_t *stt_create_object(stt_ctx_t *stt, stt_obj_t *type);

/* create a generic */
stt_obj_t *stt_create_generic(stt_ctx_t *stt);

/* create a method */
stt_obj_t *stt_create_method(stt_ctx_t *stt);

/* create a dispatch state */
stt_obj_t *stt_create_dispatch(stt_ctx_t *stt);

/* create a varref */
stt_obj_t *stt_create_varref(stt_ctx_t *stt, ptrdiff_t offset);

/* add a weak reference place to object */
void stt_weak_reference(stt_obj_t *host, stt_serial_t serial, stt_obj_t **ptr);

/* remove stt_weakptr_t references */
STT_INLINE void stt_weak_unref(stt_obj_t **o);

/* check weakptrs, clean up invalid entries
 * returns number of entries removed */
int stt_weak_check(stt_obj_t *obj);



/****** generic object functions ******/


/* get next type id, create random hash seed */
int stt_gettypeid(stt_ctx_t *stt);

/* get type id for symbol. -1 if no such thing */
int stt_symbol_to_typeid(stt_ctx_t *stt, stt_obj_t *sym);

/* return type id of object */
int stt_object_typeid(stt_obj_t *obj);

/* print object to a FP (for debugging) */
stt_obj_t *stt_fp_printobj(FILE *fp, stt_varctx_t *ctx, const stt_obj_t *obj);

/* print description/value of an object */
stt_obj_t *stt_printobj(stt_varctx_t *ctx, const stt_obj_t *obj);

/* free inner data of object */
void stt_freedata(stt_obj_t *obj);

/* deep copy object */
void stt_deepcopy(stt_ctx_t *stt, stt_obj_t *tgt, stt_obj_t *obj);

/* returns true if two objects are stt_isoform(all fields are equal) */
int stt_isoform(const stt_obj_t *a, const stt_obj_t *b);



/****** CLOS support *****/


/* calculate and update internal data of class */
stt_obj_t *stt_class_calculate(stt_ctx_t *stt, stt_obj_t *cl);

/* calculate and update data for method */
stt_obj_t *stt_method_calculate(stt_varctx_t *ctx, stt_obj_t *name, stt_obj_t *method);

/* update next pointers for all methods in a generic */
void stt_update_generic_nexts(stt_ctx_t *stt, stt_obj_t *generic);

/* compare specificity of two methods
 * returns
 *  <0   if a is less specific than b
 *  0    if a has the same signature as b
 *  >0   if a is more specific than b
 */
int stt_compare_specificity(stt_ctx_t *stt, const stt_obj_t *a, const stt_obj_t *b);

/* calculate class precedence */
void stt_build_precedence(stt_ctx_t *stt);

/* returns true if A class is a generalisation of, or equal to, B class */
int stt_is_generalisation(const stt_obj_t *a, const stt_obj_t *b);

/* returns true if B method is a possible next-method of A method. */
int stt_is_next_method(stt_ctx_t *stt, const stt_obj_t *a, const stt_obj_t *b);

/* expand/zero dispatch table */
void stt_expand_dispatch(stt_obj_t *generic);

/* checks if a method is applicable to an argument array.
 * (last element NULL)
 * returns first argument in the array unaccounted for
 */
stt_obj_t **stt_is_applicable(stt_ctx_t *stt, const stt_obj_t *method, stt_obj_t **args);

/* do method dispatch
 * argeval should be an array of stt_obj_t * (last one NULL) of pre-evaluated arguments.
 * returns the STT_ST_METHOD the call should use, or NULL
 * if ignore_around is 1, around methods are ignored
 */
stt_obj_t *stt_method_dispatch(stt_ctx_t *stt, stt_obj_t *generic, stt_obj_t **argeval, const int ignore_around);

/* write row into dispatch table
 * finds first empty row starting from <row>
 * returns that row, after filling it in.
 */
int stt_write_dispatch(stt_obj_t *generic, int row, int hashval, stt_obj_t **argeval, stt_obj_t *method, const int ignore_around);


/****** dictionaries ******/




/* get hash value for an object */
unsigned long stt_hashfunc(const stt_obj_t *obj);

/* put value into dictionary (helper function) */
int stt_dictput_(stt_dictentry_t *data, unsigned int mask, unsigned long keyhash, stt_obj_t *key, stt_obj_t *value);

/* put value into dictionary */
void stt_dictput(stt_ctx_t *stt, stt_obj_t *dict, unsigned long keyhash, stt_obj_t *key, stt_obj_t *value);

/* rehash dictionary */
void stt_dictrehash(stt_ctx_t *stt, stt_obj_t *dict);

/* get value from dictionary */
stt_obj_t *stt_dictget(const stt_obj_t *dict, const stt_obj_t *key);

/* delete item from dictionary */
int stt_dictdel(stt_obj_t *dict, const stt_obj_t *key);



/****** streams ******/

/* returns the value of *output-stream* */
stt_obj_t *stt_get_output_stream(stt_varctx_t *ctx);

/* returns the value of *input-stream* */
stt_obj_t *stt_get_input_stream(stt_varctx_t *ctx);

/* write a character to a stream */
stt_obj_t *stt_stream_write_char(stt_varctx_t *ctx, stt_obj_t *stream, int ch);

/* unread a character into a stream */
stt_obj_t *stt_stream_unread_char(stt_ctx_t *stt, stt_obj_t *stream, const unsigned char ch);

/* read a character form a stream */
stt_obj_t *stt_stream_read_char(stt_varctx_t *ctx, stt_obj_t *stream);

/* read a character from *input-stream* */
stt_obj_t *stt_stream_stdin_read(stt_varctx_t *ctx);

/* unread a character into *input-stream* */
stt_obj_t *stt_stream_stdin_unread(stt_varctx_t *ctx, const unsigned char ch);

/* write a character to *output-stream* */
stt_obj_t *stt_stream_stdout_write(stt_varctx_t *ctx, int ch);

/* read buffer from instream */
stt_obj_t *stt_stream_readbuf(stt_varctx_t *ctx, unsigned char *buf, size_t bytes, size_t *readlen);

/* dump buffer to standard output stream */
stt_obj_t *stt_stream_dumpbuf(stt_varctx_t *ctx, unsigned char *buf, size_t len);

/* printf to the standard output stream */
stt_obj_t *stt_stream_printf(stt_varctx_t *ctx, const char *format, ...);



/****** utility ******/


/* creates a new string builder */
stt_strbuild_t stt_strbuild_new(void);

/* append a character */
void stt_strbuild_addch(stt_strbuild_t *token, int ch);

/* append len characters from a string */
void stt_strbuild_addstr(stt_strbuild_t *token, const char *str, size_t len);

/* define a new cookie type (returns id) */
int stt_new_cookie_type(void);

/* find occurrence of one blob in another
 * returns offset, or -1
 * dir 0: from the left
 * dir 1: from the right
 */
int stt_blob_seek(unsigned char *needle, size_t needle_size,
	unsigned char *haystack, size_t haystack_size, int dir);

/* quote an object for passing as parameter.
 * returns a list: (quote OBJ)
 */
stt_obj_t *stt_quoteobj(stt_ctx_t *stt, stt_obj_t *obj);

#endif
